# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

import json
from dataclasses import dataclass
from typing import TYPE_CHECKING

import sqlalchemy as sa
from twisted.internet import defer

from buildbot.db import NULL
from buildbot.db import base
from buildbot.util import epoch2datetime
from buildbot.util.twisted import async_to_deferred
from buildbot.warnings import warn_deprecated

if TYPE_CHECKING:
    import datetime
    from typing import Sequence

    from buildbot.data.resultspec import ResultSpec
    from buildbot.db.sourcestamps import SourceStampModel


@dataclass
class BuildModel:
    id: int
    number: int
    builderid: int
    buildrequestid: int
    workerid: int | None
    masterid: int
    started_at: datetime.datetime
    complete_at: datetime.datetime | None
    locks_duration_s: int | None
    state_string: str
    results: int | None

    # For backward compatibility
    def __getitem__(self, key: str):
        warn_deprecated(
            '4.1.0',
            (
                'BuildsConnectorComponent getBuild, '
                'getBuildByNumber, getPrevSuccessfulBuild, '
                'getBuildsForChange, getBuilds, '
                '_getRecentBuilds, and _getBuild '
                'no longer return Build as dictionnaries. '
                'Usage of [] accessor is deprecated: please access the member directly'
            ),
        )

        if hasattr(self, key):
            return getattr(self, key)

        raise KeyError(key)


class BuildsConnectorComponent(base.DBConnectorComponent):
    def _getBuild(self, whereclause) -> defer.Deferred[BuildModel | None]:
        def thd(conn) -> BuildModel | None:
            q = self.db.model.builds.select()
            if whereclause is not None:
                q = q.where(whereclause)
            res = conn.execute(q)
            row = res.fetchone()

            rv = None
            if row:
                rv = self._model_from_row(row)
            res.close()
            return rv

        return self.db.pool.do(thd)

    def getBuild(self, buildid: int) -> defer.Deferred[BuildModel | None]:
        return self._getBuild(self.db.model.builds.c.id == buildid)

    def getBuildByNumber(self, builderid: int, number: int) -> defer.Deferred[BuildModel | None]:
        return self._getBuild(
            (self.db.model.builds.c.builderid == builderid)
            & (self.db.model.builds.c.number == number)
        )

    def _getRecentBuilds(self, whereclause, offset=0, limit=1) -> defer.Deferred[list[BuildModel]]:
        def thd(conn) -> list[BuildModel]:
            tbl = self.db.model.builds

            q = tbl.select()
            if whereclause is not None:
                q = q.where(
                    whereclause,
                )

            q = (
                q.order_by(
                    sa.desc(tbl.c.complete_at),
                )
                .offset(offset)
                .limit(limit)
            )

            res = conn.execute(q)
            return list(self._model_from_row(row) for row in res.fetchall())

        return self.db.pool.do(thd)

    @defer.inlineCallbacks
    def getPrevSuccessfulBuild(
        self, builderid: int, number: int, ssBuild: Sequence[SourceStampModel]
    ):
        gssfb = self.master.db.sourcestamps.getSourceStampsForBuild
        rv = None
        tbl = self.db.model.builds
        offset = 0
        increment = 1000
        matchssBuild = {(ss.repository, ss.branch, ss.codebase) for ss in ssBuild}
        while rv is None:
            # Get some recent successful builds on the same builder
            prevBuilds = yield self._getRecentBuilds(
                whereclause=(
                    (tbl.c.builderid == builderid) & (tbl.c.number < number) & (tbl.c.results == 0)
                ),
                offset=offset,
                limit=increment,
            )
            if not prevBuilds:
                break
            for prevBuild in prevBuilds:
                prevssBuild = {
                    (ss.repository, ss.branch, ss.codebase) for ss in (yield gssfb(prevBuild.id))
                }
                if prevssBuild == matchssBuild:
                    # A successful build with the same
                    # repository/branch/codebase was found !
                    rv = prevBuild
                    break
            offset += increment

        return rv

    def getBuildsForChange(self, changeid: int) -> defer.Deferred[list[BuildModel]]:
        assert changeid > 0

        def thd(conn) -> list[BuildModel]:
            # Get builds for the change
            changes_tbl = self.db.model.changes
            bsets_tbl = self.db.model.buildsets
            bsss_tbl = self.db.model.buildset_sourcestamps
            reqs_tbl = self.db.model.buildrequests
            builds_tbl = self.db.model.builds

            from_clause = changes_tbl.join(
                bsss_tbl, changes_tbl.c.sourcestampid == bsss_tbl.c.sourcestampid
            )
            from_clause = from_clause.join(bsets_tbl, bsss_tbl.c.buildsetid == bsets_tbl.c.id)
            from_clause = from_clause.join(reqs_tbl, bsets_tbl.c.id == reqs_tbl.c.buildsetid)
            from_clause = from_clause.join(builds_tbl, reqs_tbl.c.id == builds_tbl.c.buildrequestid)

            q = (
                sa.select(builds_tbl)
                .select_from(from_clause)
                .where(changes_tbl.c.changeid == changeid)
            )
            res = conn.execute(q)
            return [self._model_from_row(row) for row in res.fetchall()]

        return self.db.pool.do(thd)

    def getBuilds(
        self,
        builderid: int | None = None,
        buildrequestid: int | None = None,
        workerid: int | None = None,
        complete: bool | None = None,
        resultSpec: ResultSpec | None = None,
    ) -> defer.Deferred[list[BuildModel]]:
        def thd(conn) -> list[BuildModel]:
            tbl = self.db.model.builds
            q = tbl.select()
            if builderid is not None:
                q = q.where(tbl.c.builderid == builderid)
            if buildrequestid is not None:
                q = q.where(tbl.c.buildrequestid == buildrequestid)
            if workerid is not None:
                q = q.where(tbl.c.workerid == workerid)
            if complete is not None:
                if complete:
                    q = q.where(tbl.c.complete_at != NULL)
                else:
                    q = q.where(tbl.c.complete_at == NULL)

            if resultSpec is not None:
                return resultSpec.thd_execute(conn, q, self._model_from_row)

            res = conn.execute(q)
            return [self._model_from_row(row) for row in res.fetchall()]

        return self.db.pool.do(thd)

    @async_to_deferred
    async def get_triggered_builds(self, buildid: int) -> list[BuildModel]:
        def thd(conn) -> list[BuildModel]:
            j = self.db.model.buildsets
            j = j.join(self.db.model.buildrequests)
            j = j.join(self.db.model.builds)

            q = (
                sa.select(self.db.model.builds)
                .select_from(j)
                .where(self.db.model.buildsets.c.parent_buildid == buildid)
            )

            res = conn.execute(q)
            return [self._model_from_row(row) for row in res.fetchall()]

        return await self.db.pool.do(thd)

    # returns a Deferred that returns a value
    def addBuild(
        self, builderid, buildrequestid, workerid, masterid, state_string, _race_hook=None
    ):
        started_at = int(self.master.reactor.seconds())

        def thd(conn):
            tbl = self.db.model.builds
            # get the highest current number
            r = conn.execute(
                sa.select(sa.func.max(tbl.c.number)).where(tbl.c.builderid == builderid)
            )
            number = r.scalar()
            new_number = 1 if number is None else number + 1
            # insert until we are successful..
            while True:
                if _race_hook:
                    _race_hook(conn)

                try:
                    r = conn.execute(
                        self.db.model.builds.insert(),
                        {
                            "number": new_number,
                            "builderid": builderid,
                            "buildrequestid": buildrequestid,
                            "workerid": workerid,
                            "masterid": masterid,
                            "started_at": started_at,
                            "complete_at": None,
                            "locks_duration_s": 0,
                            "state_string": state_string,
                        },
                    )
                    conn.commit()
                except (sa.exc.IntegrityError, sa.exc.ProgrammingError) as e:
                    conn.rollback()
                    # pg 9.5 gives this error which makes it pass some build
                    # numbers
                    if 'duplicate key value violates unique constraint "builds_pkey"' not in str(e):
                        new_number += 1
                    continue
                return r.inserted_primary_key[0], new_number

        return self.db.pool.do(thd)

    # returns a Deferred that returns None
    def setBuildStateString(self, buildid, state_string):
        def thd(conn):
            tbl = self.db.model.builds

            q = tbl.update().where(tbl.c.id == buildid)
            conn.execute(q.values(state_string=state_string))

        return self.db.pool.do_with_transaction(thd)

    # returns a Deferred that returns None
    def finishBuild(self, buildid, results):
        def thd(conn):
            tbl = self.db.model.builds
            q = tbl.update().where(tbl.c.id == buildid)
            conn.execute(q.values(complete_at=int(self.master.reactor.seconds()), results=results))

        return self.db.pool.do_with_transaction(thd)

    # returns a Deferred that returns a value
    def getBuildProperties(self, bid, resultSpec=None):
        def thd(conn):
            bp_tbl = self.db.model.build_properties
            q = sa.select(
                bp_tbl.c.name,
                bp_tbl.c.value,
                bp_tbl.c.source,
            ).where(bp_tbl.c.buildid == bid)
            props = []
            if resultSpec is not None:
                data = resultSpec.thd_execute(conn, q, lambda x: x)
            else:
                data = conn.execute(q)
            for row in data:
                prop = (json.loads(row.value), row.source)
                props.append((row.name, prop))
            return dict(props)

        return self.db.pool.do(thd)

    @defer.inlineCallbacks
    def setBuildProperty(self, bid, name, value, source):
        """A kind of create_or_update, that's between one or two queries per
        call"""

        def thd(conn):
            bp_tbl = self.db.model.build_properties
            self.checkLength(bp_tbl.c.name, name)
            self.checkLength(bp_tbl.c.source, source)
            whereclause = sa.and_(bp_tbl.c.buildid == bid, bp_tbl.c.name == name)
            q = sa.select(bp_tbl.c.value, bp_tbl.c.source).where(whereclause)
            prop = conn.execute(q).fetchone()
            value_js = json.dumps(value)
            if prop is None:
                conn.execute(
                    bp_tbl.insert(),
                    {"buildid": bid, "name": name, "value": value_js, "source": source},
                )
            elif (prop.value != value_js) or (prop.source != source):
                conn.execute(
                    bp_tbl.update().where(whereclause), {"value": value_js, "source": source}
                )

        yield self.db.pool.do_with_transaction(thd)

    @defer.inlineCallbacks
    def add_build_locks_duration(self, buildid, duration_s):
        def thd(conn):
            builds_tbl = self.db.model.builds
            conn.execute(
                builds_tbl.update()
                .where(builds_tbl.c.id == buildid)
                .values(locks_duration_s=builds_tbl.c.locks_duration_s + duration_s)
            )

        yield self.db.pool.do_with_transaction(thd)

    def _model_from_row(self, row):
        return BuildModel(
            id=row.id,
            number=row.number,
            builderid=row.builderid,
            buildrequestid=row.buildrequestid,
            workerid=row.workerid,
            masterid=row.masterid,
            started_at=epoch2datetime(row.started_at),
            complete_at=epoch2datetime(row.complete_at),
            locks_duration_s=row.locks_duration_s,
            state_string=row.state_string,
            results=row.results,
        )



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

import json
from typing import TYPE_CHECKING

from buildbot.data import connector
from buildbot.data import resultspec
from buildbot.test.util import validation
from buildbot.util import service
from buildbot.util.twisted import async_to_deferred

if TYPE_CHECKING:
    import datetime


class FakeUpdates(service.AsyncService):
    # unlike "real" update methods, all of the fake methods are here in a
    # single class.

    def __init__(self, testcase, data):
        self.testcase = testcase
        self.data = data

        # test cases should assert the values here:
        self.changesAdded = []  # Changes are numbered starting at 1.

    def assertProperties(self, sourced, properties):
        self.testcase.assertIsInstance(properties, dict)
        for k, v in properties.items():
            self.testcase.assertIsInstance(k, str)
            if sourced:
                self.testcase.assertIsInstance(v, tuple)
                self.testcase.assertEqual(len(v), 2)
                propval, propsrc = v
                self.testcase.assertIsInstance(propsrc, str)
            else:
                propval = v
            try:
                json.dumps(propval)
            except (TypeError, ValueError):
                self.testcase.fail(f"value for {k} is not JSON-able")

    # update methods

    def addChange(
        self,
        files=None,
        comments=None,
        author=None,
        committer=None,
        revision=None,
        when_timestamp=None,
        branch=None,
        category=None,
        revlink='',
        properties=None,
        repository='',
        codebase=None,
        project='',
        src=None,
    ):
        # double-check args, types, etc.
        if files is not None:
            self.testcase.assertIsInstance(files, list)
            map(lambda f: self.testcase.assertIsInstance(f, str), files)
        self.testcase.assertIsInstance(comments, (type(None), str))
        self.testcase.assertIsInstance(author, (type(None), str))
        self.testcase.assertIsInstance(committer, (type(None), str))
        self.testcase.assertIsInstance(revision, (type(None), str))
        self.testcase.assertIsInstance(when_timestamp, (type(None), int))
        self.testcase.assertIsInstance(branch, (type(None), str))
        self.testcase.assertIsInstance(revlink, (type(None), str))
        if properties is not None:
            self.assertProperties(sourced=False, properties=properties)
        self.testcase.assertIsInstance(repository, str)
        self.testcase.assertIsInstance(codebase, (type(None), str))
        self.testcase.assertIsInstance(project, str)
        self.testcase.assertIsInstance(src, (type(None), str))

        self.changesAdded.append({
            'files': files,
            'comments': comments,
            'author': author,
            'committer': committer,
            'revision': revision,
            'when_timestamp': when_timestamp,
            'branch': branch,
            'category': category,
            'revlink': revlink,
            'properties': properties.copy() if properties is not None else None,
            'repository': repository,
            'codebase': codebase,
            'project': project,
            'src': src,
        })
        return self.data.updates.addChange(
            files=files,
            comments=comments,
            author=author,
            committer=committer,
            revision=revision,
            when_timestamp=when_timestamp,
            branch=branch,
            category=category,
            revlink=revlink,
            properties=properties,
            repository=repository,
            codebase=codebase,
            project=project,
            src=src,
        )

    def masterActive(self, name, masterid):
        self.testcase.assertIsInstance(name, str)
        self.testcase.assertIsInstance(masterid, int)
        if masterid:
            self.testcase.assertEqual(masterid, 1)
        return self.data.updates.masterActive(name, masterid)

    def masterStopped(self, name, masterid):
        self.testcase.assertIsInstance(name, str)
        self.testcase.assertEqual(masterid, 1)
        return self.data.updates.masterStopped(name, masterid)

    def expireMasters(self, forceHouseKeeping=False):
        return self.data.updates.expireMasters(forceHouseKeeping=forceHouseKeeping)

    @async_to_deferred
    async def addBuildset(
        self,
        waited_for,
        scheduler=None,
        sourcestamps=None,
        reason='',
        properties=None,
        builderids=None,
        external_idstring=None,
        rebuilt_buildid=None,
        parent_buildid=None,
        parent_relationship=None,
        priority=0,
    ) -> tuple[int, dict[int, int]]:
        self.testcase.assertIsInstance(scheduler, str)
        self.testcase.assertIsInstance(sourcestamps, (type(None), list))
        if sourcestamps is not None:
            for ss in sourcestamps:
                if not isinstance(ss, int) and not isinstance(ss, dict):
                    self.testcase.fail(f"{ss} ({type(ss)}) is not an integer or a dictionary")
        self.testcase.assertIsInstance(reason, str)
        if properties is not None:
            self.assertProperties(sourced=True, properties=properties)
        self.testcase.assertIsInstance(builderids, (type(None), list))
        self.testcase.assertIsInstance(external_idstring, (type(None), str))

        return await self.data.updates.addBuildset(
            waited_for,
            scheduler=scheduler,
            sourcestamps=sourcestamps,
            reason=reason,
            properties=properties,
            builderids=builderids,
            external_idstring=external_idstring,
            rebuilt_buildid=rebuilt_buildid,
            parent_buildid=parent_buildid,
            parent_relationship=parent_relationship,
            priority=priority,
        )

    def maybeBuildsetComplete(self, bsid):
        return self.data.updates.maybeBuildsetComplete(bsid)

    @async_to_deferred
    async def claimBuildRequests(self, brids, claimed_at=None) -> bool:
        validation.verifyType(
            self.testcase, 'brids', brids, validation.ListValidator(validation.IntValidator())
        )
        validation.verifyType(
            self.testcase,
            'claimed_at',
            claimed_at,
            validation.NoneOk(validation.DateTimeValidator()),
        )
        return await self.data.updates.claimBuildRequests(brids, claimed_at=claimed_at)

    @async_to_deferred
    async def unclaimBuildRequests(self, brids) -> None:
        validation.verifyType(
            self.testcase, 'brids', brids, validation.ListValidator(validation.IntValidator())
        )
        return await self.data.updates.unclaimBuildRequests(brids)

    def completeBuildRequests(self, brids, results, complete_at=None):
        validation.verifyType(
            self.testcase, 'brids', brids, validation.ListValidator(validation.IntValidator())
        )
        validation.verifyType(self.testcase, 'results', results, validation.IntValidator())
        validation.verifyType(
            self.testcase,
            'complete_at',
            complete_at,
            validation.NoneOk(validation.DateTimeValidator()),
        )
        return self.data.updates.completeBuildRequests(brids, results, complete_at=complete_at)

    def rebuildBuildrequest(self, buildrequest):
        return self.data.updates.rebuildBuildrequest(buildrequest)

    @async_to_deferred
    async def update_project_info(
        self,
        projectid,
        slug,
        description,
        description_format,
        description_html,
    ) -> None:
        return await self.data.updates.update_project_info(
            projectid, slug, description, description_format, description_html
        )

    def find_project_id(self, name: str, auto_create: bool = True):
        validation.verifyType(self.testcase, 'project name', name, validation.StringValidator())
        validation.verifyType(
            self.testcase, 'auto_create', auto_create, validation.BooleanValidator()
        )
        return self.data.updates.find_project_id(name)

    @async_to_deferred
    async def add_commit(
        self,
        *,
        codebaseid: int,
        author: str,
        committer: str | None = None,
        files: list[str] | None = None,
        comments: str,
        when_timestamp: datetime.datetime,
        revision: str,
        parent_commitid: int | None = None,
    ) -> None:
        validation.verifyType(self.testcase, 'codebaseid', codebaseid, validation.IntValidator())
        validation.verifyType(self.testcase, 'author', author, validation.StringValidator())
        validation.verifyType(
            self.testcase, 'committer', committer, validation.NoneOk(validation.StringValidator())
        )
        validation.verifyType(
            self.testcase,
            'files',
            files,
            validation.NoneOk(validation.StringListValidator()),
        )
        validation.verifyType(self.testcase, 'comments', comments, validation.StringValidator())
        validation.verifyType(
            self.testcase, 'when_timestamp', when_timestamp, validation.IntValidator()
        )
        validation.verifyType(self.testcase, 'revision', revision, validation.StringValidator())
        validation.verifyType(
            self.testcase,
            'parent_commitid',
            parent_commitid,
            validation.NoneOk(validation.IntValidator()),
        )

        return await self.data.updates.add_commit(
            codebaseid=codebaseid,
            author=author,
            committer=committer,
            files=files,
            comments=comments,
            when_timestamp=when_timestamp,
            revision=revision,
            parent_commitid=parent_commitid,
        )

    @async_to_deferred
    async def update_branch(
        self,
        *,
        codebaseid: int,
        name: str,
        commitid: int | None = None,
        last_timestamp: datetime.datetime,
    ) -> None:
        validation.verifyType(self.testcase, 'codebaseid', codebaseid, validation.IntValidator())
        validation.verifyType(self.testcase, 'name', name, validation.StringValidator())
        validation.verifyType(
            self.testcase, 'commitid', commitid, validation.NoneOk(validation.IntValidator())
        )
        validation.verifyType(
            self.testcase, 'last_timestamp', last_timestamp, validation.IntValidator()
        )

        return await self.data.updates.update_branch(
            codebaseid=codebaseid,
            name=name,
            commitid=commitid,
            last_timestamp=last_timestamp,
        )

    @async_to_deferred
    async def update_codebase_info(
        self,
        *,
        codebaseid: int,
        projectid: int,
        slug: str,
    ) -> None:
        validation.verifyType(self.testcase, 'codebaseid', codebaseid, validation.IntValidator())
        validation.verifyType(self.testcase, 'projectid', projectid, validation.IntValidator())
        validation.verifyType(self.testcase, 'slug', slug, validation.StringValidator())

        await self.data.updates.update_codebase_info(
            codebaseid=codebaseid, projectid=projectid, slug=slug
        )

    def find_codebase_id(self, *, projectid: int, name: str, auto_create: bool = True):
        validation.verifyType(self.testcase, 'project id', projectid, validation.IntValidator())
        validation.verifyType(self.testcase, 'codebase name', name, validation.StringValidator())
        return self.data.updates.find_codebase_id(
            projectid=projectid, name=name, auto_create=auto_create
        )

    def updateBuilderList(self, masterid, builderNames):
        self.testcase.assertEqual(masterid, self.master.masterid)
        for n in builderNames:
            self.testcase.assertIsInstance(n, str)
        self.builderNames = builderNames
        return self.data.updates.updateBuilderList(masterid, builderNames)

    @async_to_deferred
    async def updateBuilderInfo(
        self, builderid, description, description_format, description_html, projectid, tags
    ) -> None:
        await self.data.updates.updateBuilderInfo(
            builderid, description, description_format, description_html, projectid, tags
        )

    def findSchedulerId(self, name):
        return self.data.updates.findSchedulerId(name)

    def findChangeSourceId(self, name):
        validation.verifyType(
            self.testcase, 'changesource name', name, validation.StringValidator()
        )
        return self.data.updates.findChangeSourceId(name)

    def findBuilderId(self, name):
        validation.verifyType(self.testcase, 'builder name', name, validation.StringValidator())
        return self.data.updates.findBuilderId(name)

    def trySetSchedulerMaster(self, schedulerid, masterid):
        return self.data.updates.trySetSchedulerMaster(schedulerid, masterid)

    def trySetChangeSourceMaster(self, changesourceid, masterid):
        return self.data.updates.trySetChangeSourceMaster(changesourceid, masterid)

    def addBuild(self, builderid, buildrequestid, workerid):
        validation.verifyType(self.testcase, 'builderid', builderid, validation.IntValidator())
        validation.verifyType(
            self.testcase, 'buildrequestid', buildrequestid, validation.IntValidator()
        )
        validation.verifyType(self.testcase, 'workerid', workerid, validation.IntValidator())
        return self.data.updates.addBuild(builderid, buildrequestid, workerid)

    def generateNewBuildEvent(self, buildid):
        validation.verifyType(self.testcase, 'buildid', buildid, validation.IntValidator())
        return self.data.updates.generateNewBuildEvent(buildid)

    def setBuildStateString(self, buildid, state_string):
        validation.verifyType(self.testcase, 'buildid', buildid, validation.IntValidator())
        validation.verifyType(
            self.testcase, 'state_string', state_string, validation.StringValidator()
        )
        return self.data.updates.setBuildStateString(buildid, state_string)

    def add_build_locks_duration(self, buildid, duration_s):
        validation.verifyType(self.testcase, 'buildid', buildid, validation.IntValidator())
        validation.verifyType(self.testcase, 'duration_s', duration_s, validation.IntValidator())
        return self.data.updates.add_build_locks_duration(buildid, duration_s)

    def finishBuild(self, buildid, results):
        validation.verifyType(self.testcase, 'buildid', buildid, validation.IntValidator())
        validation.verifyType(self.testcase, 'results', results, validation.IntValidator())
        return self.data.updates.finishBuild(buildid, results)

    def setBuildProperty(self, buildid, name, value, source):
        validation.verifyType(self.testcase, 'buildid', buildid, validation.IntValidator())
        validation.verifyType(self.testcase, 'name', name, validation.StringValidator())
        try:
            json.dumps(value)
        except (TypeError, ValueError):
            self.testcase.fail(f"Value for {name} is not JSON-able")
        validation.verifyType(self.testcase, 'source', source, validation.StringValidator())
        return self.data.updates.setBuildProperty(buildid, name, value, source)

    def setBuildProperties(self, buildid, properties) -> None:
        return self.data.updates.setBuildProperties(buildid, properties)

    def addStep(self, buildid, name):
        validation.verifyType(self.testcase, 'buildid', buildid, validation.IntValidator())
        validation.verifyType(self.testcase, 'name', name, validation.IdentifierValidator(50))
        return self.data.updates.addStep(buildid, name)

    def addStepURL(self, stepid, name, url):
        validation.verifyType(self.testcase, 'stepid', stepid, validation.IntValidator())
        validation.verifyType(self.testcase, 'name', name, validation.StringValidator())
        validation.verifyType(self.testcase, 'url', url, validation.StringValidator())
        return self.data.updates.addStepURL(stepid, name, url)

    def startStep(self, stepid, started_at=None, locks_acquired=False):
        validation.verifyType(self.testcase, 'stepid', stepid, validation.IntValidator())
        validation.verifyType(
            self.testcase, "started_at", started_at, validation.NoneOk(validation.IntValidator())
        )
        validation.verifyType(
            self.testcase, "locks_acquired", locks_acquired, validation.BooleanValidator()
        )
        return self.data.updates.startStep(
            stepid, started_at=started_at, locks_acquired=locks_acquired
        )

    def set_step_locks_acquired_at(self, stepid, locks_acquired_at=None):
        validation.verifyType(self.testcase, 'stepid', stepid, validation.IntValidator())
        validation.verifyType(
            self.testcase,
            "locks_acquired_at",
            locks_acquired_at,
            validation.NoneOk(validation.IntValidator()),
        )
        return self.data.updates.set_step_locks_acquired_at(
            stepid, locks_acquired_at=locks_acquired_at
        )

    def setStepStateString(self, stepid, state_string):
        validation.verifyType(self.testcase, 'stepid', stepid, validation.IntValidator())
        validation.verifyType(
            self.testcase, 'state_string', state_string, validation.StringValidator()
        )
        return self.data.updates.setStepStateString(stepid, state_string)

    def finishStep(self, stepid, results, hidden):
        validation.verifyType(self.testcase, 'stepid', stepid, validation.IntValidator())
        validation.verifyType(self.testcase, 'results', results, validation.IntValidator())
        validation.verifyType(self.testcase, 'hidden', hidden, validation.BooleanValidator())
        return self.data.updates.finishStep(stepid, results, hidden)

    def addLog(self, stepid, name, type):
        validation.verifyType(self.testcase, 'stepid', stepid, validation.IntValidator())
        validation.verifyType(self.testcase, 'name', name, validation.StringValidator())
        validation.verifyType(self.testcase, 'type', type, validation.IdentifierValidator(1))
        return self.data.updates.addLog(stepid, name, type)

    def finishLog(self, logid):
        validation.verifyType(self.testcase, 'logid', logid, validation.IntValidator())
        return self.data.updates.finishLog(logid)

    def compressLog(self, logid):
        validation.verifyType(self.testcase, 'logid', logid, validation.IntValidator())
        return self.data.updates.compressLog(logid)

    def appendLog(self, logid, content):
        validation.verifyType(self.testcase, 'logid', logid, validation.IntValidator())
        validation.verifyType(self.testcase, 'content', content, validation.StringValidator())
        self.testcase.assertEqual(content[-1], '\n')
        return self.data.updates.appendLog(logid, content)

    def findWorkerId(self, name):
        validation.verifyType(
            self.testcase, 'worker name', name, validation.IdentifierValidator(50)
        )
        # this needs to actually get inserted into the db (fake or real) since
        # getWorker will get called later
        return self.data.updates.findWorkerId(name)

    def workerConnected(self, workerid, masterid, workerinfo):
        return self.data.updates.workerConnected(
            workerid=workerid, masterid=masterid, workerinfo=workerinfo
        )

    def workerConfigured(self, workerid, masterid, builderids):
        return self.data.updates.workerConfigured(
            workerid=workerid, masterid=masterid, builderids=builderids
        )

    def workerDisconnected(self, workerid, masterid):
        return self.data.updates.workerDisconnected(workerid=workerid, masterid=masterid)

    def deconfigureAllWorkersForMaster(self, masterid):
        return self.data.updates.deconfigureAllWorkersForMaster(masterid=masterid)

    def workerMissing(self, workerid, masterid, last_connection, notify):
        return self.data.updates.workerMissing(workerid, masterid, last_connection, notify)

    def schedulerEnable(self, schedulerid, v):
        return self.data.updates.schedulerEnable(schedulerid, v)

    def set_worker_paused(self, workerid, paused, pause_reason=None):
        return self.data.updates.set_worker_paused(workerid, paused, pause_reason)

    def set_worker_graceful(self, workerid, graceful):
        return self.data.updates.set_worker_graceful(workerid, graceful)

    # methods form BuildData resource
    @async_to_deferred
    async def setBuildData(self, buildid, name, value, source) -> None:
        validation.verifyType(self.testcase, 'buildid', buildid, validation.IntValidator())
        validation.verifyType(self.testcase, 'name', name, validation.StringValidator())
        validation.verifyType(self.testcase, 'value', value, validation.BinaryValidator())
        validation.verifyType(self.testcase, 'source', source, validation.StringValidator())
        return await self.data.updates.setBuildData(buildid, name, value, source)

    # methods from TestResultSet resource
    @async_to_deferred
    async def addTestResultSet(
        self, builderid, buildid, stepid, description, category, value_unit
    ) -> int:
        validation.verifyType(self.testcase, 'builderid', builderid, validation.IntValidator())
        validation.verifyType(self.testcase, 'buildid', buildid, validation.IntValidator())
        validation.verifyType(self.testcase, 'stepid', stepid, validation.IntValidator())
        validation.verifyType(
            self.testcase, 'description', description, validation.StringValidator()
        )
        validation.verifyType(self.testcase, 'category', category, validation.StringValidator())
        validation.verifyType(self.testcase, 'value_unit', value_unit, validation.StringValidator())
        return await self.data.updates.addTestResultSet(
            builderid, buildid, stepid, description, category, value_unit
        )

    @async_to_deferred
    async def completeTestResultSet(
        self, test_result_setid, tests_passed=None, tests_failed=None
    ) -> None:
        validation.verifyType(
            self.testcase, 'test_result_setid', test_result_setid, validation.IntValidator()
        )
        validation.verifyType(
            self.testcase,
            'tests_passed',
            tests_passed,
            validation.NoneOk(validation.IntValidator()),
        )
        validation.verifyType(
            self.testcase,
            'tests_failed',
            tests_failed,
            validation.NoneOk(validation.IntValidator()),
        )

        await self.data.updates.completeTestResultSet(test_result_setid, tests_passed, tests_failed)

    # methods from TestResult resource
    @async_to_deferred
    async def addTestResults(self, builderid, test_result_setid, result_values) -> None:
        await self.data.updates.addTestResults(builderid, test_result_setid, result_values)


class FakeDataConnector(service.AsyncMultiService):
    # FakeDataConnector delegates to the real DataConnector so it can get all
    # of the proper getter and consumer behavior; it overrides all of the
    # relevant updates with fake methods, though.

    def __init__(self, master, testcase):
        super().__init__()
        self.setServiceParent(master)

        # get, control and updates are delegated to a real connector,
        # after some additional assertions
        self.realConnector = connector.DataConnector()
        self.realConnector.setServiceParent(self)

        self.updates = FakeUpdates(testcase, self.realConnector)
        self.updates.setServiceParent(self)
        self.rtypes = self.realConnector.rtypes

    def _scanModule(self, mod):
        return self.realConnector._scanModule(mod)

    def getEndpoint(self, path):
        if not isinstance(path, tuple):
            raise TypeError('path must be a tuple')
        return self.realConnector.getEndpoint(path)

    def getResourceType(self, name):
        return getattr(self.rtypes, name)

    def get(self, path, filters=None, fields=None, order=None, limit=None, offset=None):
        if not isinstance(path, tuple):
            raise TypeError('path must be a tuple')
        return self.realConnector.get(
            path, filters=filters, fields=fields, order=order, limit=limit, offset=offset
        )

    def get_with_resultspec(self, path, rspec):
        if not isinstance(path, tuple):
            raise TypeError('path must be a tuple')
        if not isinstance(rspec, resultspec.ResultSpec):
            raise TypeError('rspec must be ResultSpec')
        return self.realConnector.get_with_resultspec(path, rspec)

    def control(self, action, args, path):
        if not isinstance(path, tuple):
            raise TypeError('path must be a tuple')
        return self.realConnector.control(action, args, path)

    def resultspec_from_jsonapi(self, args, entityType, is_collection):
        return self.realConnector.resultspec_from_jsonapi(args, entityType, is_collection)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING
from typing import Protocol

if TYPE_CHECKING:
    from typing import ClassVar


class CompressObjInterface(Protocol):
    def __init__(self) -> None:
        pass

    @abstractmethod
    def compress(self, data: bytes) -> bytes:
        raise NotImplementedError

    @abstractmethod
    def flush(self) -> bytes:
        raise NotImplementedError


class CompressorInterface(Protocol):
    name: ClassVar[str]
    available: ClassVar[bool] = True

    CompressObj: ClassVar[type[CompressObjInterface]]

    @staticmethod
    @abstractmethod
    def dumps(data: bytes) -> bytes:
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def read(data: bytes) -> bytes:
        raise NotImplementedError



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

import hashlib
from contextlib import contextmanager
from typing import TYPE_CHECKING

import sqlalchemy as sa
from sqlalchemy.ext import compiler
from sqlalchemy.sql.elements import BooleanClauseList
from sqlalchemy.sql.elements import ColumnElement
from sqlalchemy.sql.expression import ClauseElement
from sqlalchemy.sql.expression import Executable

if TYPE_CHECKING:
    from typing import Any
    from typing import Callable
    from typing import Sequence

    from sqlalchemy.future.engine import Connection
    from sqlalchemy.future.engine import Engine

# from http:
# www.sqlalchemy.org/docs/core/compiler.html#compiling-sub-elements-of-a-custom-expression-construct
# _execution_options per
# http://docs.sqlalchemy.org/en/rel_0_7/core/compiler.html#enabling-compiled-autocommit
# (UpdateBase requires sqlalchemy 0.7.0)


class InsertFromSelect(Executable, ClauseElement):
    _execution_options = Executable._execution_options.union({'autocommit': True})

    def __init__(self, table, select):
        self.table = table
        self.select = select


@compiler.compiles(InsertFromSelect)
def _visit_insert_from_select(element, compiler, **kw):
    return (
        f"INSERT INTO {compiler.process(element.table, asfrom=True)} "
        f"{compiler.process(element.select)}"
    )


def sa_version():
    if hasattr(sa, '__version__'):

        def tryint(s):
            try:
                return int(s)
            except (ValueError, TypeError):
                return -1

        return tuple(map(tryint, sa.__version__.split('.')))
    return (0, 0, 0)  # "it's old"


def Table(*args, **kwargs):
    """Wrap table creation to add any necessary dialect-specific options"""
    # work around the case where a database was created for us with
    # a non-utf8 character set (mysql's default)
    kwargs['mysql_character_set'] = 'utf8'
    return sa.Table(*args, **kwargs)


@contextmanager
def withoutSqliteForeignKeys(connection: Connection):
    if connection.engine.dialect.name != 'sqlite':
        yield
        return

    res = connection.exec_driver_sql('pragma foreign_keys')
    r = res.fetchone()
    assert r
    foreign_keys_enabled = r[0]
    res.close()
    if not foreign_keys_enabled:
        yield
        return

    # This context is not re-entrant. Ensure it.
    assert not getattr(connection.engine, 'fk_disabled', False)
    connection.fk_disabled = True  # type: ignore[attr-defined]
    connection.exec_driver_sql('pragma foreign_keys=OFF')
    try:
        yield
    finally:
        connection.fk_disabled = False  # type: ignore[attr-defined]
        connection.exec_driver_sql('pragma foreign_keys=ON')


def get_sqlite_version():
    import sqlite3

    return sqlite3.sqlite_version_info


def get_upsert_method(engine: Engine | None):
    if engine is None:
        return _upsert_default

    # https://sqlite.org/lang_upsert.html
    if engine.dialect.name == 'sqlite' and get_sqlite_version() > (3, 24, 0):
        return _upsert_sqlite
    if engine.dialect.name == 'postgresql':
        return _upsert_postgresql
    if engine.dialect.name == 'mysql':
        return _upsert_mysql

    return _upsert_default


def _upsert_sqlite(
    connection: Connection,
    table: sa.Table,
    *,
    where_values: Sequence[tuple[sa.Column, Any]],
    update_values: Sequence[tuple[sa.Column, Any]],
    _race_hook: Callable[[Connection], None] | None = None,
):
    from sqlalchemy.dialects.sqlite import insert  # pylint: disable=import-outside-toplevel

    _upsert_on_conflict_do_update(
        insert,
        connection,
        table,
        where_values=where_values,
        update_values=update_values,
        _race_hook=_race_hook,
    )


def _upsert_postgresql(
    connection: Connection,
    table: sa.Table,
    *,
    where_values: Sequence[tuple[sa.Column, Any]],
    update_values: Sequence[tuple[sa.Column, Any]],
    _race_hook: Callable[[Connection], None] | None = None,
):
    from sqlalchemy.dialects.postgresql import insert  # pylint: disable=import-outside-toplevel

    _upsert_on_conflict_do_update(
        insert,
        connection,
        table,
        where_values=where_values,
        update_values=update_values,
        _race_hook=_race_hook,
    )


def _upsert_on_conflict_do_update(
    insert: Any,
    connection: Connection,
    table: sa.Table,
    *,
    where_values: Sequence[tuple[sa.Column, Any]],
    update_values: Sequence[tuple[sa.Column, Any]],
    _race_hook: Callable[[Connection], None] | None = None,
):
    if _race_hook is not None:
        _race_hook(connection)

    insert_stmt = insert(table).values(
        **_column_value_kwargs(where_values),
        **_column_value_kwargs(update_values),
    )
    do_update_stmt = insert_stmt.on_conflict_do_update(
        index_elements=[c for (c, _) in where_values],
        index_where=_column_values_where_clause(where_values),
        set_=dict(update_values),
    )
    connection.execute(do_update_stmt)


def _upsert_mysql(
    connection: Connection,
    table: sa.Table,
    *,
    where_values: Sequence[tuple[sa.Column, Any]],
    update_values: Sequence[tuple[sa.Column, Any]],
    _race_hook: Callable[[Connection], None] | None = None,
):
    from sqlalchemy.dialects.mysql import insert  # pylint: disable=import-outside-toplevel

    if _race_hook is not None:
        _race_hook(connection)

    update_kwargs = _column_value_kwargs(update_values)
    insert_stmt = insert(table).values(
        **_column_value_kwargs(where_values),
        **update_kwargs,
    )
    on_duplicate_key_stmt = insert_stmt.on_duplicate_key_update(
        **update_kwargs,
    )
    connection.execute(on_duplicate_key_stmt)


def _upsert_default(
    connection: Connection,
    table: sa.Table,
    *,
    where_values: Sequence[tuple[sa.Column, Any]],
    update_values: Sequence[tuple[sa.Column, Any]],
    _race_hook: Callable[[Connection], None] | None = None,
):
    q = table.update()
    if where_values:
        q = q.where(_column_values_where_clause(where_values))
    res = connection.execute(q.values(*update_values))
    if res.rowcount > 0:
        return
    # the update hit 0 rows, so try inserting a new one

    if _race_hook is not None:
        _race_hook(connection)

    connection.execute(
        table.insert().values(
            **_column_value_kwargs(where_values),
            **_column_value_kwargs(update_values),
        )
    )


def _column_value_kwargs(values: Sequence[tuple[sa.Column, Any]]) -> dict[str, Any]:
    return {c.name: v for (c, v) in values}


def _column_values_where_clause(values: Sequence[tuple[sa.Column, Any]]) -> ColumnElement[bool]:
    return BooleanClauseList.and_(*[c == v for (c, v) in values])


def hash_columns(*args):
    def encode(x):
        if x is None:
            return b'\xf5'
        elif isinstance(x, str):
            return x.encode('utf-8')
        return str(x).encode('utf-8')

    return hashlib.sha1(b'\0'.join(map(encode, args))).hexdigest()



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


from __future__ import annotations

import json
import re
from typing import TYPE_CHECKING
from typing import Any

from dateutil.parser import parse as dateparse
from twisted.internet.defer import inlineCallbacks
from twisted.python import log

from buildbot.process.properties import Properties
from buildbot.util import bytes2unicode
from buildbot.www.hooks.base import BaseHookHandler

if TYPE_CHECKING:
    from twisted.web.server import Request

    from buildbot.util.twisted import InlineCallbacksType

_HEADER_EVENT = b'X-Gitlab-Event'
_HEADER_GITLAB_TOKEN = b'X-Gitlab-Token'


class GitLabHandler(BaseHookHandler):
    def _process_change(
        self,
        payload: dict[str, Any],
        user: str,
        repo: str,
        repo_url: str,
        event: str,
        codebase: str | None = None,
    ) -> list[dict[str, Any]]:
        """
        Consumes the JSON as a python object and actually starts the build.

        :arguments:
            payload
                Python Object that represents the JSON sent by GitLab Service
                Hook.
        """
        changes: list[dict[str, Any]] = []
        refname = payload['ref']
        # project name from http headers is empty for me, so get it from repository/name
        project = payload['repository']['name']

        # We only care about regular heads or tags
        match = re.match(r"^refs/(heads|tags)/(.+)$", refname)
        if not match:
            log.msg(f"Ignoring refname `{refname}': Not a branch")
            return changes

        branch = match.group(2)
        if payload.get('deleted'):
            log.msg(f"Branch `{branch}' deleted, ignoring")
            return changes

        for commit in payload['commits']:
            if not commit.get('distinct', True):
                log.msg(f"Commit `{commit['id']}` is a non-distinct commit, ignoring...")
                continue

            files = []
            for kind in ('added', 'modified', 'removed'):
                files.extend(commit.get(kind, []))

            when_timestamp = dateparse(commit['timestamp'])

            log.msg(f"New revision: {commit['id'][:8]}")

            change = {
                'author': f"{commit['author']['name']} <{commit['author']['email']}>",
                'files': files,
                'comments': commit['message'],
                'revision': commit['id'],
                'when_timestamp': when_timestamp,
                'branch': branch,
                'revlink': commit['url'],
                'repository': repo_url,
                'project': project,
                'category': event,
                'properties': {
                    'event': event,
                },
            }

            if codebase is not None:
                change['codebase'] = codebase

            changes.append(change)

        return changes

    def _process_merge_request_change(
        self, payload: dict[str, Any], event: str, codebase: str | None = None
    ) -> list[dict[str, Any]]:
        """
        Consumes the merge_request JSON as a python object and turn it into a buildbot change.

        :arguments:
            payload
                Python Object that represents the JSON sent by GitLab Service
                Hook.
        """
        attrs = payload['object_attributes']
        commit = attrs['last_commit']
        when_timestamp = dateparse(commit['timestamp'])
        # @todo provide and document a way to choose between http and ssh url
        repo_url = attrs['target']['git_http_url']
        # project name from http headers is empty for me, so get it from
        # object_attributes/target/name
        project = attrs['target']['name']

        # Filter out uninteresting events
        state = attrs['state']
        if re.match('^(closed|merged|approved)$', state):
            log.msg(f"GitLab MR#{attrs['iid']}: Ignoring because state is {state}")
            return []
        action = attrs['action']
        if not re.match('^(open|reopen)$', action) and not (
            action == "update" and "oldrev" in attrs
        ):
            log.msg(
                f"GitLab MR#{attrs['iid']}: Ignoring because action {action} was not open or "
                "reopen or an update that added code"
            )
            return []

        changes = [
            {
                'author': f"{commit['author']['name']} <{commit['author']['email']}>",
                'files': [],  # @todo use rest API
                'comments': f"MR#{attrs['iid']}: {attrs['title']}\n\n{attrs['description']}",
                'revision': commit['id'],
                'when_timestamp': when_timestamp,
                'branch': attrs['target_branch'],
                'repository': repo_url,
                'project': project,
                'category': event,
                'revlink': attrs['url'],
                'properties': {
                    'source_branch': attrs['source_branch'],
                    'source_project_id': attrs['source_project_id'],
                    'source_repository': attrs['source']['git_http_url'],
                    'source_git_ssh_url': attrs['source']['git_ssh_url'],
                    'target_branch': attrs['target_branch'],
                    'target_project_id': attrs['target_project_id'],
                    'target_repository': attrs['target']['git_http_url'],
                    'target_git_ssh_url': attrs['target']['git_ssh_url'],
                    'event': event,
                },
            }
        ]
        if codebase is not None:
            changes[0]['codebase'] = codebase
        return changes

    def _process_note_addition_to_merge_request(
        self, payload: dict[str, Any], event: str, codebase: str | None = None
    ) -> list[dict[str, Any]]:
        """
        Consumes a note event JSON as a python object and turn it into a buildbot change.

        :arguments:
            payload
                Python Object that represents the JSON sent by GitLab Service
                Hook.

        Comments in merge_requests are send as note events by the API
        """
        attrs = payload['object_attributes']

        # handle only note events coming from merge_requests
        # this can be direct comments or comments added to a changeset of the MR
        #
        # editing a comment does NOT lead to an event at all
        if 'merge_request' not in payload:
            log.msg(f"Found note event (id {attrs['id']}) without corresponding MR - ignore")
            return []

        # change handling is very similar to the method above, but
        commit = payload['merge_request']['last_commit']
        when_timestamp = dateparse(commit['timestamp'])
        # @todo provide and document a way to choose between http and ssh url
        repo_url = payload['merge_request']['target']['git_http_url']
        # project name from http headers is empty for me, so get it from
        # object_attributes/target/name
        mr = payload['merge_request']
        project = mr['target']['name']

        log.msg(f"Found notes on MR#{mr['iid']}: {attrs['note']}")
        changes = [
            {
                'author': f"{commit['author']['name']} <{commit['author']['email']}>",
                'files': [],  # not provided by rest API
                'comments': f"MR#{mr['iid']}: {mr['title']}\n\n{mr['description']}",
                'revision': commit['id'],
                'when_timestamp': when_timestamp,
                'branch': mr['target_branch'],
                'repository': repo_url,
                'project': project,
                'category': event,
                'revlink': mr['url'],
                'properties': {
                    'source_branch': mr['source_branch'],
                    'source_project_id': mr['source_project_id'],
                    'source_repository': mr['source']['git_http_url'],
                    'source_git_ssh_url': mr['source']['git_ssh_url'],
                    'target_branch': mr['target_branch'],
                    'target_project_id': mr['target_project_id'],
                    'target_repository': mr['target']['git_http_url'],
                    'target_git_ssh_url': mr['target']['git_ssh_url'],
                    'event': event,
                    'comments': attrs['note'],
                },
            }
        ]
        if codebase is not None:
            changes[0]['codebase'] = codebase
        return changes

    @inlineCallbacks
    def getChanges(self, request: Request) -> InlineCallbacksType[tuple[list[dict[str, Any]], str]]:  # type: ignore[override]
        """
        Reponds only to POST events and starts the build process

        :arguments:
            request
                the http request object
        """
        expected_secret = isinstance(self.options, dict) and self.options.get('secret')
        if expected_secret:
            received_secret = bytes2unicode(request.getHeader(_HEADER_GITLAB_TOKEN))

            p = Properties()
            p.master = self.master
            expected_secret_value = yield p.render(expected_secret)

            if received_secret != expected_secret_value:
                raise ValueError("Invalid secret")
        try:
            assert request.content is not None
            content = request.content.read()
            payload = json.loads(bytes2unicode(content))
        except Exception as e:
            raise ValueError("Error loading JSON: " + str(e)) from e

        assert request.args is not None

        event_type = bytes2unicode(request.getHeader(_HEADER_EVENT))
        # newer version of gitlab have a object_kind parameter,
        # which allows not to use the http header
        event_type = payload.get('object_kind', event_type)
        codebase = bytes2unicode(request.args.get(b'codebase', [None])[0])
        if event_type in ("push", "tag_push", "Push Hook"):
            user = payload['user_name']
            repo = payload['repository']['name']
            repo_url = payload['repository']['url']
            changes = self._process_change(
                payload, user, repo, repo_url, event_type, codebase=codebase
            )
        elif event_type == 'merge_request':
            changes = self._process_merge_request_change(payload, event_type, codebase=codebase)
        elif event_type == 'note':
            changes = self._process_note_addition_to_merge_request(
                payload, event_type, codebase=codebase
            )
        else:
            changes = []
        if changes:
            log.msg(f"Received {len(changes)} changes from {event_type!r} gitlab event")
        return (changes, 'git')


gitlab = GitLabHandler



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import warnings
from unittest import mock

import setuptools  # force import setuptools before any other distutils imports

from buildbot import monkeypatches
from buildbot.test.util.warnings import assertProducesWarning  # noqa: F401
from buildbot.test.util.warnings import assertProducesWarnings  # noqa: F401
from buildbot.warnings import DeprecatedApiWarning  # noqa: F401

_ = mock

# apply the same patches the buildmaster does when it starts
monkeypatches.patch_all(for_tests=True)

# enable deprecation warnings
warnings.filterwarnings('always', category=DeprecationWarning)

_ = setuptools  # force use for pylint

# This is where we load deprecated module-level APIs to ignore warning produced by importing them.
# After the deprecated API has been removed, leave at least one instance of the import in a
# commented state as reference.

# with assertProducesWarnings(DeprecatedApiWarning,
#                             messages_patterns=[
#                                 r" buildbot\.status\.base has been deprecated",
#                             ]):
#     import buildbot.status.base as _

# All deprecated modules should be loaded, consider future warnings in tests as errors.
# In order to not pollute the test outputs,
# warnings in tests shall be forcefully tested with assertProducesWarning,
# or shutdown using the warning module
warnings.filterwarnings('error')
# if buildbot_worker is installed in pip install -e mode, then the docker directory will
# match "import docker", and produce a warning.
# We just suppress this warning instead of doing silly workaround.
warnings.filterwarnings(
    'ignore', "Not importing directory.*docker': missing __init__.py", category=ImportWarning
)

# autobahn is not updated for Twisted 22.04 and newer
warnings.filterwarnings(
    "ignore", "twisted.web.resource.NoResource was deprecated in", category=DeprecationWarning
)

# When using Python 3.12, this generates some dependent package
warnings.filterwarnings(
    'ignore',
    r"datetime.datetime.utcnow\(\) is deprecated and scheduled for "
    r"removal in a future version. Use timezone-aware objects to represent "
    r"datetimes in UTC: datetime.datetime.now\(datetime.UTC\).",
    category=DeprecationWarning,
)

# Python3.12 generates deprecation warnings like:
# "This process (pid=6558) is multi-threaded, use of fork() may lead to deadlocks in the child."
# Tracked in https://github.com/buildbot/buildbot/issues/7276
warnings.filterwarnings(
    "ignore",
    r"This process \(pid=\d+\) is multi-threaded, use of fork\(\) may lead "
    r"to deadlocks in the child\.",
    category=DeprecationWarning,
)

# Warnings comes from attr 24.1.0 because of automat
warnings.filterwarnings(
    "ignore",
    r"The `hash` argument is deprecated in favor of `unsafe_hash` "
    r"and will be removed in or after August 2025\.",
    category=DeprecationWarning,
)

warnings.filterwarnings(
    "ignore",
    r"twisted.web.resource._UnsafeErrorPage.__init__ was deprecated in "
    r"Twisted 22.10.0; please use Use twisted.web.pages.errorPage instead, "
    r"which properly escapes HTML. instead",
    category=DeprecationWarning,
)

warnings.filterwarnings(
    "ignore",
    r"twisted.web.resource._UnsafeNoResource.__init__ was deprecated in "
    r"Twisted 22.10.0; please use Use twisted.web.pages.notFound instead, "
    r"which properly escapes HTML. instead",
    category=DeprecationWarning,
)

warnings.filterwarnings(
    "ignore",
    r"twisted.web.resource._UnsafeForbiddenResource.__init__ was deprecated in "
    r"Twisted 22.10.0; please use Use twisted.web.pages.forbidden instead, "
    r"which properly escapes HTML. instead",
    category=DeprecationWarning,
)

# Warnings comes form ldap3. See https://github.com/cannatag/ldap3/issues/1159
# builtins.DeprecationWarning: typeMap is deprecated. Please use TYPE_MAP instead.
warnings.filterwarnings(
    "ignore",
    r".* is deprecated\. Please use .* instead\.",
    category=DeprecationWarning,
)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import re
from typing import ClassVar
from typing import Sequence

from twisted.internet import defer

from buildbot import config
from buildbot.data import resultspec
from buildbot.util.service import BuildbotService
from buildbot.util.ssfilter import SourceStampFilter
from buildbot.util.ssfilter import extract_filter_values


class _OldBuildFilterSet:
    def __init__(self):
        self._by_builder = {}

    def add_filter(self, builders, filter):
        assert builders is not None

        for builder in builders:
            self._by_builder.setdefault(builder, []).append(filter)

    def is_matched(self, builder_name, props):
        assert builder_name is not None

        filters = self._by_builder.get(builder_name, [])
        for filter in filters:
            if filter.is_matched(props):
                return True
        return False


class _TrackedBuildRequest:
    def __init__(self, brid, builder_name, start_time, ss_tuples):
        self.start_time = start_time
        self.builder_name = builder_name
        self.brid = brid
        self.ss_tuples = ss_tuples

    def __str__(self):
        return (
            f'_TrackedBuildRequest({self.brid}, {self.builder_name} '
            f'{self.start_time}, {self.ss_tuples})'
        )

    __repr__ = __str__


class _OldBuildrequestTracker:
    def __init__(self, reactor, filter, branch_key, on_cancel):
        self.reactor = reactor
        self.filter = filter
        self.branch_key = branch_key
        self.on_cancel = on_cancel

        # We need to track build requests by IDs so that when such build request finishes we know
        # what we no longer need to track. We also need to track build requests by source code
        # branch, so that we can cancel build requests when branch sees new commits. Branch is
        # identified by a tuple of project, codebase, repository and branch.
        #
        # Note that a single branch may run multiple builds. Also, changes are not a source for
        # build request cancelling because a change may not result in builds being started due to
        # user scheduler configuration. In such case it makes sense to let the build finish.

        # (is_build, id) -> _TrackedBuildRequest
        self.br_by_id = {}
        self.br_by_ss = {}
        self.change_time_by_ss = {}
        self._change_count_since_clean = 0

    def reconfig(self, filter, branch_key):
        self.filter = filter
        self.branch_key = branch_key

    def is_buildrequest_tracked(self, br_id):
        return br_id in self.br_by_id

    def on_new_buildrequest(self, brid, builder_name, sourcestamps):
        self._maybe_cancel_new_obsoleted_buildrequest(builder_name, sourcestamps)
        self._add_new_buildrequest(brid, builder_name, sourcestamps)

    def _add_new_buildrequest(self, brid, builder_name, sourcestamps):
        now = self.reactor.seconds()

        matched_ss = []

        for ss in sourcestamps:
            if ss['branch'] is None:
                return

            # Note that it's enough to match build by a single branch from a single codebase
            if self.filter.is_matched(builder_name, ss):
                matched_ss.append(ss)

        if not matched_ss:
            return

        ss_tuples = [
            (ss['project'], ss['codebase'], ss['repository'], self.branch_key(ss))
            for ss in matched_ss
        ]

        tracked_br = _TrackedBuildRequest(brid, builder_name, now, ss_tuples)
        self.br_by_id[brid] = tracked_br

        for ss_tuple in ss_tuples:
            br_dict = self.br_by_ss.setdefault(ss_tuple, {})
            br_dict[tracked_br.brid] = tracked_br

    def _maybe_cancel_new_obsoleted_buildrequest(self, builder_name, sourcestamps):
        for sourcestamp in sourcestamps:
            ss_tuple = (
                sourcestamp['project'],
                sourcestamp['codebase'],
                sourcestamp['repository'],
                self.branch_key(sourcestamp),
            )

            newest_change_time = self.change_time_by_ss.get(ss_tuple, None)
            if newest_change_time is None:
                # Don't cancel any buildrequests if the cancelling buildrequest does not come from
                # a change.
                continue

            br_dict = self.br_by_ss.get(ss_tuple, None)
            if br_dict is None:
                continue

            brids_to_cancel = []
            for tracked_br in list(br_dict.values()):
                if tracked_br.builder_name != builder_name:
                    continue

                if newest_change_time <= tracked_br.start_time:
                    # The existing build request is newer than the change, thus change should not
                    # be a reason to cancel it.
                    continue

                del self.br_by_id[tracked_br.brid]

                # Clear the canceled buildrequest from self.br_by_ss
                for i_ss_tuple in tracked_br.ss_tuples:
                    other_br_dict = self.br_by_ss.get(i_ss_tuple, None)
                    if other_br_dict is None:
                        raise KeyError(
                            f'{self.__class__.__name__}: Could not find running builds '
                            f'by tuple {i_ss_tuple}'
                        )

                    del other_br_dict[tracked_br.brid]
                    if not other_br_dict:
                        del self.br_by_ss[i_ss_tuple]

                brids_to_cancel.append(tracked_br.brid)

            for brid in brids_to_cancel:
                self.on_cancel(brid)

    def on_complete_buildrequest(self, brid):
        tracked_br = self.br_by_id.pop(brid, None)
        if tracked_br is None:
            return

        for ss_tuple in tracked_br.ss_tuples:
            br_dict = self.br_by_ss.get(ss_tuple, None)
            if br_dict is None:
                raise KeyError(
                    f'{self.__class__.__name__}: Could not find finished builds by tuple {ss_tuple}'
                )

            del br_dict[tracked_br.brid]
            if not br_dict:
                del self.br_by_ss[ss_tuple]

    def on_change(self, change):
        now = self.reactor.seconds()

        ss_tuple = (
            change['project'],
            change['codebase'],
            change['repository'],
            self.branch_key(change),
        )

        # Note that now is monotonically increasing
        self.change_time_by_ss[ss_tuple] = now

        self._change_count_since_clean += 1
        if self._change_count_since_clean >= 1000:
            self.change_time_by_ss = {
                ss_tuple: ss_now
                for ss_tuple, ss_now in self.change_time_by_ss.items()
                if now - ss_now < 60 * 10
            }
            self._change_count_since_clean = 0


class OldBuildCanceller(BuildbotService):
    compare_attrs: ClassVar[Sequence[str]] = (*BuildbotService.compare_attrs, 'filters')

    def checkConfig(self, name, filters, branch_key=None):
        OldBuildCanceller.check_filters(filters)

        self.name = name

        self._buildrequest_new_consumer = None
        self._buildrequest_complete_consumer = None

        self._build_tracker = None
        self._reconfiguring = False
        self._completed_buildrequests_while_reconfiguring = []

    @defer.inlineCallbacks
    def reconfigService(self, name, filters, branch_key=None):
        # While reconfiguring we acquire a list of currently pending build
        # requests and seed the build tracker with these. We need to ensure that even if some
        # builds or build requests finish during this process, the tracker gets to know about
        # the changes in correct order. In order to do that, we defer all build request completion
        # notifications to after the reconfig finishes.
        #
        # Note that old builds are cancelled according to the configuration that was live when they
        # were created, so for already tracked builds we don't need to do anything.
        self._reconfiguring = True

        if branch_key is None:
            branch_key = self._default_branch_key

        filter_set_object = OldBuildCanceller.filter_tuples_to_filter_set_object(filters)

        if self._build_tracker is None:
            self._build_tracker = _OldBuildrequestTracker(
                self.master.reactor, filter_set_object, branch_key, self._cancel_buildrequest
            )
        else:
            self._build_tracker.reconfig(filter_set_object, branch_key)

        all_running_buildrequests = yield self.master.data.get(
            ('buildrequests',), filters=[resultspec.Filter('complete', 'eq', [False])]
        )

        for breq in all_running_buildrequests:
            if self._build_tracker.is_buildrequest_tracked(breq['buildrequestid']):
                continue
            yield self._on_buildrequest_new(None, breq)

        self._reconfiguring = False

        completed_breqs = self._completed_buildrequests_while_reconfiguring
        self._completed_buildrequests_while_reconfiguring = []

        for breq in completed_breqs:
            self._build_tracker.on_complete_buildrequest(breq['buildrequestid'])

    @defer.inlineCallbacks
    def startService(self):
        yield super().startService()
        self._change_consumer = yield self.master.mq.startConsuming(
            self._on_change, ('changes', None, 'new')
        )
        self._buildrequest_new_consumer = yield self.master.mq.startConsuming(
            self._on_buildrequest_new, ('buildrequests', None, 'new')
        )
        self._buildrequest_complete_consumer = yield self.master.mq.startConsuming(
            self._on_buildrequest_complete, ('buildrequests', None, 'complete')
        )

    @defer.inlineCallbacks
    def stopService(self):
        yield self._change_consumer.stopConsuming()
        yield self._buildrequest_new_consumer.stopConsuming()
        yield self._buildrequest_complete_consumer.stopConsuming()

    @classmethod
    def check_filters(cls, filters):
        if not isinstance(filters, list):
            config.error(f'{cls.__name__}: The filters argument must be a list of tuples')

        for filter in filters:
            if (
                not isinstance(filter, tuple)
                or len(filter) != 2
                or not isinstance(filter[1], SourceStampFilter)
            ):
                config.error(
                    (
                        '{}: The filters argument must be a list of tuples each of which '
                        + 'contains builders as the first item and SourceStampFilter as '
                        + 'the second'
                    ).format(cls.__name__)
                )

            builders, _ = filter

            try:
                extract_filter_values(builders, 'builders')
            except Exception as e:
                config.error(f'{cls.__name__}: When processing filter builders: {e!s}')

    @classmethod
    def filter_tuples_to_filter_set_object(cls, filters):
        filter_set = _OldBuildFilterSet()

        for filter in filters:
            builders, ss_filter = filter
            filter_set.add_filter(extract_filter_values(builders, 'builders'), ss_filter)

        return filter_set

    def _default_branch_key(self, ss_or_change):
        branch = ss_or_change['branch']
        if branch is None:
            return None

        # On some VCS systems each iteration of a PR gets its own branch. We want to track all
        # iterations of the PR as a single unit.
        if branch.startswith('refs/changes/'):
            m = re.match(r'refs/changes/(\d+)/(\d+)/\d+', branch)
            if m is not None:
                return f'refs/changes/{m.group(1)}/{m.group(2)}'

        return branch

    def _on_change(self, key, change):
        self._build_tracker.on_change(change)

    @defer.inlineCallbacks
    def _on_buildrequest_new(self, key, breq):
        builder = yield self.master.data.get(("builders", breq['builderid']))
        buildset = yield self.master.data.get(('buildsets', breq['buildsetid']))

        self._build_tracker.on_new_buildrequest(
            breq['buildrequestid'], builder['name'], buildset['sourcestamps']
        )

    def _on_buildrequest_complete(self, key, breq):
        if self._reconfiguring:
            self._completed_buildrequests_while_reconfiguring.append(breq)
            return
        self._build_tracker.on_complete_buildrequest(breq['buildrequestid'])

    def _cancel_buildrequest(self, brid):
        self.master.data.control(
            'cancel',
            {'reason': 'Build request has been obsoleted by a newer commit'},
            ('buildrequests', str(brid)),
        )



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


import os

from twisted.internet import defer
from twisted.internet import reactor

from buildbot.test.util.decorators import flaky
from buildbot.test.util.integration import RunMasterBase


# This integration test tests that the try command line works end2end
class TryClientE2E(RunMasterBase):
    timeout = 15

    @defer.inlineCallbacks
    def setup_config(self):
        c = {}
        from buildbot.config import BuilderConfig
        from buildbot.plugins import schedulers
        from buildbot.plugins import steps
        from buildbot.process.factory import BuildFactory

        c['schedulers'] = [
            schedulers.Try_Userpass(
                name="try", builderNames=["testy"], port='tcp:0', userpass=[("alice", "pw1")]
            )
        ]
        f = BuildFactory()
        f.addStep(steps.ShellCommand(command='echo hello'))
        c['builders'] = [BuilderConfig(name="testy", workernames=["local1"], factory=f)]
        yield self.setup_master(c)

    @flaky(bugNumber=7084)
    @defer.inlineCallbacks
    def test_shell(self):
        yield self.setup_config()

        def trigger_callback():
            port = self.master.pbmanager.dispatchers['tcp:0'].port.getHost().port

            def thd():
                os.system(
                    f"buildbot try --connect=pb --master=127.0.0.1:{port} -b testy "
                    "--property=foo:bar --username=alice --passwd=pw1 --vc=none"
                )

            reactor.callInThread(thd)

        build = yield self.doForceBuild(
            wantSteps=True, triggerCallback=trigger_callback, wantLogs=True, wantProperties=True
        )
        self.assertEqual(build['buildid'], 1)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

import fnmatch
import re
from typing import TYPE_CHECKING
from typing import Any
from typing import Callable
from typing import cast

from twisted.internet import defer
from twisted.web.error import Error
from zope.interface import implementer

from buildbot.interfaces import IConfigured
from buildbot.util import unicode2bytes
from buildbot.www.authz.roles import RolesFromOwner

if TYPE_CHECKING:
    from buildbot.util.twisted import InlineCallbacksType


class Forbidden(Error):
    def __init__(self, msg: bytes) -> None:
        super().__init__(403, msg)


# fnmatch and re.match are reversed API, we cannot just rename them
def fnmatchStrMatcher(value: str, match: str) -> bool:
    return fnmatch.fnmatch(value, match)


def reStrMatcher(value: str, match: str) -> bool:
    return cast(bool, re.match(match, value))


@implementer(IConfigured)
class Authz:
    def getConfigDict(self) -> dict:
        return {}

    def __init__(
        self,
        allowRules: list[Any] | None = None,
        roleMatchers: list[Any] | None = None,
        stringsMatcher: Callable[[str, str], bool] = fnmatchStrMatcher,
    ) -> None:
        self.match = stringsMatcher
        if allowRules is None:
            allowRules = []
        if roleMatchers is None:
            roleMatchers = []
        self.allowRules = allowRules
        self.roleMatchers = [r for r in roleMatchers if not isinstance(r, RolesFromOwner)]
        self.ownerRoleMatchers = [r for r in roleMatchers if isinstance(r, RolesFromOwner)]

    def setMaster(self, master: Any) -> None:
        self.master = master
        for r in self.roleMatchers + self.ownerRoleMatchers + self.allowRules:
            r.setAuthz(self)

    def getRolesFromUser(self, userDetails: dict[str, Any]) -> set[str]:
        roles = set()
        for roleMatcher in self.roleMatchers:
            roles.update(set(roleMatcher.getRolesFromUser(userDetails)))
        return roles

    def getOwnerRolesFromUser(self, userDetails: dict[str, Any], owner: str | None) -> set[str]:
        roles = set()
        for roleMatcher in self.ownerRoleMatchers:
            roles.update(set(roleMatcher.getRolesFromUser(userDetails, owner)))
        return roles

    @defer.inlineCallbacks
    def assertUserAllowed(
        self, ep: str, action: str, options: dict[str, Any], userDetails: dict[str, Any]
    ) -> InlineCallbacksType[None]:
        roles = self.getRolesFromUser(userDetails)
        for rule in self.allowRules:
            match = yield rule.match(ep, action, options)
            if match is not None:
                # only try to get owner if there are owner Matchers
                if self.ownerRoleMatchers:
                    owner = yield match.getOwner()
                    if owner is not None:
                        roles.update(self.getOwnerRolesFromUser(userDetails, owner))
                for role in roles:
                    if self.match(role, rule.role):
                        return None

                if not rule.defaultDeny:
                    continue  # check next suitable rule if not denied

                error_msg = unicode2bytes(f"you need to have role '{rule.role}'")
                raise Forbidden(error_msg)
        return None



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


from twisted.internet import defer

from buildbot.test.util.integration import RunMasterBase


# This integration test creates a master and worker environment,
# with one builder and a shellcommand step
# meant to be a template for integration steps
class ShellMaster(RunMasterBase):
    @defer.inlineCallbacks
    def setup_config(self):
        c = {}
        from buildbot.config import BuilderConfig
        from buildbot.plugins import schedulers
        from buildbot.plugins import steps
        from buildbot.process.factory import BuildFactory

        c['schedulers'] = [schedulers.ForceScheduler(name="force", builderNames=["testy"])]

        f = BuildFactory()
        f.addStep(steps.ShellCommand(command='echo hello'))
        c['builders'] = [
            BuilderConfig(
                name="testy",
                workernames=["local1"],
                properties={
                    'virtual_builder_name': 'virtual_testy',
                    'virtual_builder_description': 'I am a virtual builder',
                    'virtual_builder_project': 'virtual_project',
                    'virtual_builder_tags': ['virtual'],
                },
                factory=f,
            )
        ]
        yield self.setup_master(c)

    @defer.inlineCallbacks
    def test_shell(self):
        yield self.setup_config()
        build = yield self.doForceBuild(wantSteps=True, wantLogs=True)
        self.assertEqual(build['buildid'], 1)
        builders = yield self.master.data.get(("builders",))
        self.assertEqual(len(builders), 2)
        self.assertEqual(
            builders[1],
            {
                'masterids': [],
                'tags': ['virtual', '_virtual_'],
                'projectid': 1,
                'description': 'I am a virtual builder',
                'description_format': None,
                'description_html': None,
                'name': 'virtual_testy',
                'builderid': 2,
            },
        )
        self.assertEqual(build['builderid'], builders[1]['builderid'])



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members
from __future__ import annotations

from fnmatch import fnmatch


class PullRequestMixin:
    external_property_whitelist: list[str] = []
    external_property_denylist: list[str] = []

    def extractProperties(self, payload):
        def flatten(properties, base, info_dict):
            for k, v in info_dict.items():
                name = ".".join([base, k])
                if name in self.external_property_denylist:
                    continue
                if isinstance(v, dict):
                    flatten(properties, name, v)
                elif any(fnmatch(name, expr) for expr in self.external_property_whitelist):
                    properties[name] = v

        properties = {}
        flatten(properties, self.property_basename, payload)
        return properties



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from unittest import mock

from twisted.internet import defer
from twisted.trial import unittest

from buildbot.mq import simple
from buildbot.test.fake import fakemaster
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.util import interfaces
from buildbot.test.util import tuplematching


class Tests(interfaces.InterfaceTests):
    def setUp(self):
        raise NotImplementedError

    def test_empty_produce(self):
        self.mq.produce(('a', 'b', 'c'), {"x": 1})
        # ..nothing happens

    def test_signature_produce(self):
        @self.assertArgSpecMatches(self.mq.produce)
        def produce(self, routingKey, data):
            pass

    def test_signature_startConsuming(self):
        @self.assertArgSpecMatches(self.mq.startConsuming)
        def startConsuming(self, callback, filter, persistent_name=None):
            pass

    @defer.inlineCallbacks
    def test_signature_stopConsuming(self):
        cons = yield self.mq.startConsuming(lambda: None, ('a',))

        @self.assertArgSpecMatches(cons.stopConsuming)
        def stopConsuming(self):
            pass

    def test_signature_waitUntilEvent(self):
        @self.assertArgSpecMatches(self.mq.waitUntilEvent)
        def waitUntilEvent(self, filter, check_callback):
            pass


class RealTests(tuplematching.TupleMatchingMixin, Tests):
    # tests that only "real" implementations will pass

    # called by the TupleMatchingMixin methods

    @defer.inlineCallbacks
    def do_test_match(self, routingKey, shouldMatch, filter):
        cb = mock.Mock()
        yield self.mq.startConsuming(cb, filter)
        self.mq.produce(routingKey, 'x')
        self.assertEqual(shouldMatch, cb.call_count == 1)
        if shouldMatch:
            cb.assert_called_once_with(routingKey, 'x')

    @defer.inlineCallbacks
    def test_stopConsuming(self):
        cb = mock.Mock()
        qref = yield self.mq.startConsuming(cb, ('abc',))
        self.mq.produce(('abc',), {"x": 1})
        qref.stopConsuming()
        self.mq.produce(('abc',), {"x": 1})
        cb.assert_called_once_with(('abc',), {"x": 1})

    @defer.inlineCallbacks
    def test_stopConsuming_twice(self):
        cb = mock.Mock()
        qref = yield self.mq.startConsuming(cb, ('abc',))
        qref.stopConsuming()
        qref.stopConsuming()
        # ..nothing bad happens

    @defer.inlineCallbacks
    def test_non_persistent(self):
        cb = mock.Mock()
        qref = yield self.mq.startConsuming(cb, ('abc',))

        cb2 = mock.Mock()
        qref2 = yield self.mq.startConsuming(cb2, ('abc',))

        qref.stopConsuming()
        self.mq.produce(('abc',), '{}')

        qref = yield self.mq.startConsuming(cb, ('abc',))
        qref.stopConsuming()
        qref2.stopConsuming()

        self.assertTrue(cb2.called)
        self.assertFalse(cb.called)

    @defer.inlineCallbacks
    def test_persistent(self):
        cb = mock.Mock()

        qref = yield self.mq.startConsuming(cb, ('abc',), persistent_name='ABC')
        qref.stopConsuming()

        self.mq.produce(('abc',), '{}')

        qref = yield self.mq.startConsuming(cb, ('abc',), persistent_name='ABC')
        qref.stopConsuming()

        self.assertTrue(cb.called)

    @defer.inlineCallbacks
    def test_waitUntilEvent_check_false(self):
        d = self.mq.waitUntilEvent(('abc',), lambda: False)
        self.assertEqual(d.called, False)
        self.mq.produce(('abc',), {"x": 1})
        self.assertEqual(d.called, True)
        res = yield d
        self.assertEqual(res, (('abc',), {"x": 1}))


class TestFakeMQ(TestReactorMixin, unittest.TestCase, Tests):
    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        self.master = yield fakemaster.make_master(self, wantMq=True)
        self.mq = self.master.mq
        self.mq.verifyMessages = False


class TestSimpleMQ(TestReactorMixin, unittest.TestCase, RealTests):
    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        self.master = yield fakemaster.make_master(self)
        self.mq = simple.SimpleMQ()
        yield self.mq.setServiceParent(self.master)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

import copy
import enum
import functools
import re
from collections import UserList
from typing import TYPE_CHECKING

from twisted.internet import defer

from buildbot.data import exceptions
from buildbot.util.twisted import async_to_deferred
from buildbot.warnings import warn_deprecated

if TYPE_CHECKING:
    from typing import Any
    from typing import Literal

    from buildbot.data import types
    from buildbot.data.resultspec import ResultSpec
    from buildbot.db.builders import BuilderModel
    from buildbot.db.builds import BuildModel
    from buildbot.db.logs import LogModel
    from buildbot.db.steps import StepModel
    from buildbot.db.workers import WorkerModel
    from buildbot.master import BuildMaster


class EndpointKind(enum.Enum):
    SINGLE = 1
    COLLECTION = 2
    RAW = 3
    RAW_INLINE = 4


class ResourceType:
    name: str | None = None
    plural: str | None = None
    endpoints: list[type[Endpoint]] = []
    eventPathPatterns: list[str] | str = []
    entityType: types.Type | None = None

    def __init__(self, master: BuildMaster):
        self.master = master
        self.compileEventPathPatterns()

    def compileEventPathPatterns(self):
        # We'll run a single format, to get the final event path tuple
        pathPatterns = self.eventPathPatterns
        if isinstance(pathPatterns, str):
            pathPatterns = pathPatterns.split()
            warn_deprecated(
                '4.3.0',
                'ResourceType.eventPathPatterns as a multiline string is deprecated. Use '
                'eventPathPatterns as a list of strings instead.',
            )
        identifiers = re.compile(r':([^/]*)')
        for i, pp in enumerate(pathPatterns):
            pp = identifiers.sub(r'{\1}', pp)
            if pp.startswith("/"):
                pp = pp[1:]
            pathPatterns[i] = pp
        self.eventPaths = pathPatterns

    @functools.lru_cache(1)  # noqa: B019
    def getEndpoints(self):
        endpoints = self.endpoints[:]
        for i, ep in enumerate(endpoints):
            if not issubclass(ep, Endpoint):
                raise TypeError("Not an Endpoint subclass")
            endpoints[i] = ep(self, self.master)
        return endpoints

    @functools.lru_cache(1)  # noqa: B019
    def getDefaultEndpoint(self):
        for ep in self.getEndpoints():
            if ep.kind != EndpointKind.COLLECTION:
                return ep
        return None

    @functools.lru_cache(1)  # noqa: B019
    def getCollectionEndpoint(self):
        for ep in self.getEndpoints():
            if ep.kind == EndpointKind.COLLECTION or ep.isPseudoCollection:
                return ep
        return None

    @staticmethod
    def sanitizeMessage(msg):
        msg = copy.deepcopy(msg)
        return msg

    def produceEvent(self, msg, event):
        if msg is not None:
            msg = self.sanitizeMessage(msg)
            for path in self.eventPaths:
                path = path.format(**msg)
                routingKey = (*tuple(path.split("/")), event)
                self.master.mq.produce(routingKey, msg)


class SubResource:
    def __init__(self, rtype):
        self.rtype = rtype
        self.endpoints = {}
        for endpoint in rtype.endpoints:
            if endpoint.kind == EndpointKind.COLLECTION:
                self.endpoints[rtype.plural] = endpoint
            else:
                self.endpoints[rtype.name] = endpoint


class Endpoint:
    pathPatterns: list[str] | str = []
    rootLinkName: str | None = None
    isPseudoCollection = False
    kind = EndpointKind.SINGLE
    parentMapping: dict[str, str] = {}

    def __init__(self, rtype, master: BuildMaster):
        self.rtype = rtype
        self.master = master

    def get(self, resultSpec: ResultSpec, kwargs: dict[str, Any]):
        raise NotImplementedError

    async def stream(self, resultSpec: ResultSpec, kwargs: dict[str, Any]):
        """
        This is a prototype interface method for internal use.
        There could be breaking changes to it.
        Use at your own risks.
        """
        raise NotImplementedError

    def control(self, action, args, kwargs):
        # we convert the action into a mixedCase method name
        action_method = getattr(self, "action" + action.capitalize(), None)
        if action_method is None:
            raise exceptions.InvalidControlException(f"action: {action} is not supported")
        return action_method(args, kwargs)

    def __repr__(self):
        if isinstance(self.pathPatterns, str):
            self.pathPatterns = self.pathPatterns.split()
        return "endpoint for " + ",".join(self.pathPatterns)


class NestedBuildDataRetriever:
    """
    Efficiently retrieves data about various entities without repeating same queries over and over.
    The following arg keys are supported:
        - stepid
        - step_name
        - step_number
        - buildid
        - build_number
        - builderid
        - buildername
        - logid
        - log_slug
    """

    __slots__ = (
        'args',
        'build_dict',
        'builder_dict',
        'log_dict',
        'master',
        'step_dict',
        'worker_dict',
    )

    def __init__(self, master, args) -> None:
        self.master = master
        self.args = args
        # False is used as special value as "not set". None is used as "not exists". This solves
        # the problem of multiple database queries in case entity does not exist.
        self.step_dict: StepModel | None | Literal[False] = False
        self.build_dict: BuildModel | None | Literal[False] = False
        self.builder_dict: BuilderModel | None | Literal[False] = False
        self.log_dict: LogModel | None | Literal[False] = False
        self.worker_dict: WorkerModel | None | Literal[False] = False

    @async_to_deferred
    async def get_step_dict(self) -> StepModel | None:
        if self.step_dict is not False:
            return self.step_dict

        if 'stepid' in self.args:
            step_dict = self.step_dict = await self.master.db.steps.getStep(
                stepid=self.args['stepid']
            )
            return step_dict

        if 'step_name' in self.args or 'step_number' in self.args:
            build_dict = await self.get_build_dict()
            if build_dict is None:
                self.step_dict = None
                return None

            step_dict = self.step_dict = await self.master.db.steps.getStep(
                buildid=build_dict.id,
                number=self.args.get('step_number'),
                name=self.args.get('step_name'),
            )
            return step_dict

        # fallback when there's only indirect information
        if 'logid' in self.args:
            log_dict = await self.get_log_dict()
            if log_dict is not None:
                step_dict = self.step_dict = await self.master.db.steps.getStep(
                    stepid=log_dict.stepid
                )
                return step_dict

        self.step_dict = None
        return self.step_dict

    @async_to_deferred
    async def get_build_dict(self) -> BuildModel | None:
        if self.build_dict is not False:
            return self.build_dict

        if 'buildid' in self.args:
            build_dict = self.build_dict = await self.master.db.builds.getBuild(
                self.args['buildid']
            )
            return build_dict

        if 'build_number' in self.args:
            builder_id = await self.get_builder_id()

            if builder_id is None:
                self.build_dict = None
                return None

            build_dict = self.build_dict = await self.master.db.builds.getBuildByNumber(
                builderid=builder_id, number=self.args['build_number']
            )
            return build_dict

        # fallback when there's only indirect information
        step_dict = await self.get_step_dict()
        if step_dict is not None:
            build_dict = self.build_dict = await self.master.db.builds.getBuild(step_dict.buildid)
            return build_dict

        self.build_dict = None
        return None

    @async_to_deferred
    async def get_build_id(self):
        if 'buildid' in self.args:
            return self.args['buildid']

        build_dict = await self.get_build_dict()
        if build_dict is None:
            return None
        return build_dict.id

    @async_to_deferred
    async def get_builder_dict(self) -> BuilderModel | None:
        if self.builder_dict is not False:
            return self.builder_dict

        if 'builderid' in self.args:
            builder_dict = self.builder_dict = await self.master.db.builders.getBuilder(
                self.args['builderid']
            )
            return builder_dict

        if 'buildername' in self.args:
            builder_id = await self.master.db.builders.findBuilderId(
                self.args['buildername'], autoCreate=False
            )
            builder_dict = None
            if builder_id is not None:
                builder_dict = await self.master.db.builders.getBuilder(builder_id)
            self.builder_dict = builder_dict
            return builder_dict

        # fallback when there's only indirect information
        build_dict = await self.get_build_dict()
        if build_dict is not None:
            builder_dict = self.builder_dict = await self.master.db.builders.getBuilder(
                build_dict.builderid
            )
            return builder_dict

        self.builder_dict = None
        return None

    @async_to_deferred
    async def get_builder_id(self) -> int | None:
        if 'builderid' in self.args:
            return self.args['builderid']

        builder_dict = await self.get_builder_dict()
        if builder_dict is None:
            return None
        return builder_dict.id

    @async_to_deferred
    async def get_log_dict(self) -> LogModel | None:
        if self.log_dict is not False:
            return self.log_dict

        if 'logid' in self.args:
            log_dict = self.log_dict = await self.master.db.logs.getLog(self.args['logid'])
            return log_dict

        step_dict = await self.get_step_dict()
        if step_dict is None:
            self.log_dict = None
            return None
        log_dict = self.log_dict = await self.master.db.logs.getLogBySlug(
            step_dict.id, self.args.get('log_slug')
        )
        return log_dict

    @async_to_deferred
    async def get_log_id(self):
        if 'logid' in self.args:
            return self.args['logid']

        log_dict = await self.get_log_dict()
        if log_dict is None:
            return None
        return log_dict.id

    @async_to_deferred
    async def get_worker_dict(self) -> WorkerModel | None:
        if self.worker_dict is not False:
            return self.worker_dict

        build_dict = await self.get_build_dict()
        if build_dict is not None:
            workerid = build_dict.workerid
            if workerid is not None:
                worker_dict = self.worker_dict = await self.master.db.workers.getWorker(
                    workerid=workerid
                )
                return worker_dict

        self.worker_dict = None
        return None


class BuildNestingMixin:
    """
    A mixin for methods to decipher the many ways a various entities can be specified.
    """

    @defer.inlineCallbacks
    def getBuildid(self, kwargs):
        retriever = NestedBuildDataRetriever(self.master, kwargs)
        return (yield retriever.get_build_id())

    @defer.inlineCallbacks
    def getBuilderId(self, kwargs):
        retriever = NestedBuildDataRetriever(self.master, kwargs)
        return (yield retriever.get_builder_id())

    # returns Deferred that yields a number
    def get_project_id(self, kwargs):
        if "projectname" in kwargs:
            return self.master.db.projects.find_project_id(kwargs["projectname"], auto_create=False)
        return defer.succeed(kwargs["projectid"])


class ListResult(UserList):
    __slots__ = ['limit', 'offset', 'total']

    def __init__(self, values, offset=None, total=None, limit=None):
        super().__init__(values)

        # if set, this is the index in the overall results of the first element of
        # this list
        self.offset = offset

        # if set, this is the total number of results
        self.total = total

        # if set, this is the limit, either from the user or the implementation
        self.limit = limit

    def __repr__(self):
        return (
            f"ListResult({self.data!r}, offset={self.offset!r}, "
            f"total={self.total!r}, limit={self.limit!r})"
        )

    def __eq__(self, other):
        if isinstance(other, ListResult):
            return (
                self.data == other.data
                and self.offset == other.offset
                and self.total == other.total
                and self.limit == other.limit
            )
        return (
            self.data == other
            and self.offset is None
            and self.limit is None
            and (self.total is None or self.total == len(other))
        )

    def __ne__(self, other):
        return not self == other


def updateMethod(func):
    """Decorate this resourceType instance as an update method, made available
    at master.data.updates.$funcname"""
    func.isUpdateMethod = True
    return func



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

from typing import TYPE_CHECKING

from twisted.internet import defer

from buildbot.data import base
from buildbot.data import types

if TYPE_CHECKING:
    from buildbot.db.builders import BuilderModel


def _db2data(builder: BuilderModel):
    return {
        "builderid": builder.id,
        "name": builder.name,
        "masterids": builder.masterids,
        "description": builder.description,
        "description_format": builder.description_format,
        "description_html": builder.description_html,
        "projectid": builder.projectid,
        "tags": builder.tags,
    }


class BuilderEndpoint(base.BuildNestingMixin, base.Endpoint):
    kind = base.EndpointKind.SINGLE
    pathPatterns = [
        "/builders/n:builderid",
        "/builders/s:buildername",
        "/masters/n:masterid/builders/n:builderid",
    ]

    @defer.inlineCallbacks
    def get(self, resultSpec, kwargs):
        builderid = yield self.getBuilderId(kwargs)
        if builderid is None:
            return None

        builder = yield self.master.db.builders.getBuilder(builderid)
        if not builder:
            return None
        if 'masterid' in kwargs:
            if kwargs['masterid'] not in builder.masterids:
                return None
        return _db2data(builder)


class BuildersEndpoint(base.Endpoint):
    kind = base.EndpointKind.COLLECTION
    rootLinkName = 'builders'
    pathPatterns = [
        "/builders",
        "/masters/n:masterid/builders",
        "/projects/n:projectid/builders",
        "/workers/n:workerid/builders",
    ]

    @defer.inlineCallbacks
    def get(self, resultSpec, kwargs):
        bdicts = yield self.master.db.builders.getBuilders(
            masterid=kwargs.get('masterid', None),
            projectid=kwargs.get('projectid', None),
            workerid=kwargs.get('workerid', None),
        )
        return [_db2data(bd) for bd in bdicts]


class Builder(base.ResourceType):
    name = "builder"
    plural = "builders"
    endpoints = [BuilderEndpoint, BuildersEndpoint]
    eventPathPatterns = [
        "/builders/:builderid",
    ]

    class EntityType(types.Entity):
        builderid = types.Integer()
        name = types.String()
        masterids = types.List(of=types.Integer())
        description = types.NoneOk(types.String())
        description_format = types.NoneOk(types.String())
        description_html = types.NoneOk(types.String())
        projectid = types.NoneOk(types.Integer())
        tags = types.List(of=types.String())

    entityType = EntityType(name)

    @defer.inlineCallbacks
    def generateEvent(self, _id, event):
        builder = yield self.master.data.get(('builders', str(_id)))
        self.produceEvent(builder, event)

    @base.updateMethod
    def findBuilderId(self, name: str) -> defer.Deferred[int]:
        return self.master.db.builders.findBuilderId(name)

    @base.updateMethod
    @defer.inlineCallbacks
    def updateBuilderInfo(
        self,
        builderid: int,
        description: str | None,
        description_format: str | None,
        description_html: str | None,
        projectid: int,
        tags: list[int | str],
    ):
        ret = yield self.master.db.builders.updateBuilderInfo(
            builderid, description, description_format, description_html, projectid, tags
        )
        yield self.generateEvent(builderid, "update")
        return ret

    @base.updateMethod
    @defer.inlineCallbacks
    def updateBuilderList(self, masterid: int, builderNames: list[str]):
        # get the "current" list of builders for this master, so we know what
        # changes to make.  Race conditions here aren't a great worry, as this
        # is the only master inserting or deleting these records.
        builders = yield self.master.db.builders.getBuilders(masterid=masterid)

        # figure out what to remove and remove it
        builderNames_set = set(builderNames)
        for bldr in builders:
            if bldr.name not in builderNames_set:
                builderid = bldr.id
                yield self.master.db.builders.removeBuilderMaster(
                    masterid=masterid, builderid=builderid
                )
                self.master.mq.produce(
                    ('builders', str(builderid), 'stopped'),
                    {"builderid": builderid, "masterid": masterid, "name": bldr.name},
                )
            else:
                builderNames_set.remove(bldr.name)

        # now whatever's left in builderNames_set is new
        for name in builderNames_set:
            builderid = yield self.master.db.builders.findBuilderId(name)
            yield self.master.db.builders.addBuilderMaster(masterid=masterid, builderid=builderid)
            self.master.mq.produce(
                ('builders', str(builderid), 'started'),
                {"builderid": builderid, "masterid": masterid, "name": name},
            )

    # returns a Deferred that returns None
    def _masterDeactivated(self, masterid):
        # called from the masters rtype to indicate that the given master is
        # deactivated
        return self.updateBuilderList(masterid, [])



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

"""add locks_acquired_at column to steps table

Revision ID: 063
Revises: 062

"""

import sqlalchemy as sa
from alembic import op

from buildbot.util import sautils

# revision identifiers, used by Alembic.
revision = '063'
down_revision = '062'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column("steps", sa.Column("locks_acquired_at", sa.Integer, nullable=True))

    metadata = sa.MetaData()
    steps_tbl = sautils.Table(
        'steps',
        metadata,
        sa.Column("started_at", sa.Integer),
        sa.Column("locks_acquired_at", sa.Integer),
    )

    op.execute(steps_tbl.update().values({steps_tbl.c.locks_acquired_at: steps_tbl.c.started_at}))


def downgrade():
    op.drop_column("steps", "locks_acquired_at")



#!/usr/bin/env python
#
#                        Buildbot CVS Mail
#
# This script was derrived from syncmail,
# Copyright (c) 2002-2006 Barry Warsaw, Fred Drake, and contributors
#
# http://cvs-syncmail.cvs.sourceforge.net
#
# The script was re-written with the sole pupose of providing updates to
# Buildbot master by Andy Howell
#
# Options handling done right by djmitche


import optparse
import os
import re
import smtplib
import socket
import sys
import textwrap
import time
from email.utils import formataddr
from io import StringIO

"""
    -t
    --testing
      Construct message and send to stdout for testing

The rest of the command line arguments are:

    %%{sVv}
        CVS %%{sVv} loginfo expansion.  When invoked by CVS, this will be a single
        string containing the files that are changing.

"""
__version__ = '$Revision: 1.3 $'


try:
    import pwd
except ImportError:
    # pwd is not available on Windows..
    pwd = None

COMMASPACE = ', '

PROGRAM = sys.argv[0]


class SmtplibMock:
    """I stand in for smtplib for testing purposes."""

    class SMTP:
        """I stand in for smtplib.SMTP connection for testing purposes.
        I copy the message to stdout.
        """

        def close(self):
            pass

        def connect(self, mailhost, mailport):
            pass

        def sendmail(self, address, email, msg):
            sys.stdout.write(msg)


rfc822_specials_re = re.compile(r'[\(\)<>@,;:\\\"\.\[\]]')


def quotename(name):
    if name and rfc822_specials_re.search(name):
        return '"{}"'.format(name.replace('"', '\\"'))
    else:
        return name


def send_mail(options):
    # Create the smtp connection to the localhost
    conn = options.smtplib.SMTP()
    conn.connect(options.mailhost, options.mailport)
    if pwd:
        pwinfo = pwd.getpwuid(os.getuid())
        user = pwinfo[0]
        name = pwinfo[4]
    else:
        user = 'cvs'
        name = 'CVS'

    domain = options.fromhost
    if not domain:
        # getfqdn is not good for use in unit tests
        if options.amTesting:
            domain = 'testing.com'
        else:
            domain = socket.getfqdn()
    address = f'{user}@{domain}'
    s = StringIO()
    datestamp = time.strftime('%a, %d %b %Y %H:%M:%S +0000', time.gmtime(time.time()))
    fileList = ' '.join(map(str, options.files))

    vars = {
        'author': formataddr((name, address)),
        'email': options.email,
        'subject': f'cvs update for project {options.project}',
        'version': __version__,
        'date': datestamp,
    }
    print(
        '''\
From: {author}
To: {email}'''.format(**vars),
        file=s,
    )
    if options.replyto:
        print(f'Reply-To: {options.replyto}', file=s)
    print(
        '''\
Subject: {subject}
Date: {date}
X-Mailer: Python buildbot-cvs-mail {version}
'''.format(**vars),
        file=s,
    )
    print(f'Cvsmode: {options.cvsmode}', file=s)
    print(f'Category: {options.category}', file=s)
    print(f'CVSROOT: {options.cvsroot}', file=s)
    print(f'Files: {fileList}', file=s)
    if options.path:
        print(f'Path: {options.path}', file=s)
    print(f'Project: {options.project}', file=s)
    cvs_input = sys.stdin.read()

    # On Python 2, sys.stdin.read() returns bytes, but
    # on Python 3, it returns unicode str.
    if isinstance(cvs_input, bytes):
        cvs_input = cvs_input.decode("utf-8")

    s.write(cvs_input)
    print('', file=s)
    conn.sendmail(address, options.email, s.getvalue())
    conn.close()


def fork_and_send_mail(options):
    # cannot wait for child process or that will cause parent to retain cvs
    # lock for too long.  Urg!
    if not os.fork():
        # in the child
        # give up the lock you cvs thang!
        time.sleep(2)
        send_mail(options)
        os._exit(0)


description = """
This script is used to provide email notifications of changes to the CVS
repository to a buildbot master.  It is invoked via a CVS loginfo file (see
$CVSROOT/CVSROOT/loginfo).  See the Buildbot manual for more information.
"""
usage = "%prog [options] %{sVv}"
parser = optparse.OptionParser(
    description=description, usage=usage, add_help_option=True, version=__version__
)

parser.add_option(
    "-C",
    "--category",
    dest='category',
    metavar="CAT",
    help=textwrap.dedent("""\
            Category for change. This becomes the Change.category attribute, which
            can be used within the buildmaster to filter changes.
            """),
)
parser.add_option(
    "-c",
    "--cvsroot",
    dest='cvsroot',
    metavar="PATH",
    help=textwrap.dedent("""\
            CVSROOT for use by buildbot workers to checkout code.
            This becomes the Change.repository attribute.
            Exmaple: :ext:myhost:/cvsroot
            """),
)
parser.add_option(
    "-e",
    "--email",
    dest='email',
    metavar="EMAIL",
    help=textwrap.dedent("""\
            Email address of the buildbot.
            """),
)
parser.add_option(
    "-f",
    "--fromhost",
    dest='fromhost',
    metavar="HOST",
    help=textwrap.dedent("""\
            The hostname that email messages appear to be coming from.  The From:
            header of the outgoing message will look like user@hostname.  By
            default, hostname is the machine's fully qualified domain name.
            """),
)
parser.add_option(
    "-m",
    "--mailhost",
    dest='mailhost',
    metavar="HOST",
    default="localhost",
    help=textwrap.dedent("""\
            The hostname of an available SMTP server.  The default is
            'localhost'.
            """),
)
parser.add_option(
    "--mailport",
    dest='mailport',
    metavar="PORT",
    default=25,
    type="int",
    help=textwrap.dedent("""\
            The port number of SMTP server.  The default is '25'.
            """),
)
parser.add_option(
    "-q",
    "--quiet",
    dest='verbose',
    action="store_false",
    default=True,
    help=textwrap.dedent("""\
            Don't print as much status to stdout.
            """),
)
parser.add_option(
    "-p",
    "--path",
    dest='path',
    metavar="PATH",
    help=textwrap.dedent("""\
            The path for the files in this update. This comes from the %p parameter
            in loginfo for CVS version 1.12.x. Do not use this for CVS version 1.11.x
            """),
)
parser.add_option(
    "-P",
    "--project",
    dest='project',
    metavar="PROJ",
    help=textwrap.dedent("""\
            The project for the source. Often set to the CVS module being modified. This becomes
            the Change.project attribute.
            """),
)
parser.add_option(
    "-R",
    "--reply-to",
    dest='replyto',
    metavar="ADDR",
    help=textwrap.dedent("""\
            Add a "Reply-To: ADDR" header to the email message.
            """),
)
parser.add_option("-t", "--testing", action="store_true", dest="amTesting", default=False)
parser.set_defaults(smtplib=smtplib)


def get_options():
    options, args = parser.parse_args()

    # rest of command line are the files.
    options.files = args
    if options.path is None:
        options.cvsmode = '1.11'
    else:
        options.cvsmode = '1.12'

    if options.cvsroot is None:
        parser.error('--cvsroot is required')
    if options.email is None:
        parser.error('--email is required')

    # set up for unit tests
    if options.amTesting:
        options.verbose = 0
        options.smtplib = SmtplibMock

    return options


# scan args for options


def main():
    options = get_options()

    if options.verbose:
        print(f'Mailing {options.email}...')
        print('Generating notification message...')
    if options.amTesting:
        send_mail(options)
    else:
        fork_and_send_mail(options)

    if options.verbose:
        print('Generating notification message... done.')
    return 0


if __name__ == '__main__':
    ret = main()
    sys.exit(ret)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

import contextlib
import os
from threading import Lock
from typing import TYPE_CHECKING
from typing import Generic
from typing import TypeVar

from buildbot.db.compression.protocol import CompressObjInterface
from buildbot.db.compression.protocol import CompressorInterface

if TYPE_CHECKING:
    from typing import Callable
    from typing import ClassVar
    from typing import Generator


_T = TypeVar('_T')

try:
    import zstandard

    HAS_ZSTD = True
except ImportError:
    HAS_ZSTD = False


class _Pool(Generic[_T]):
    """zstandard ZstdCompressor/ZstdDecompressor are provide better performance when re-used, but are not thread-safe"""

    def __init__(self, item_ctor: Callable[[], _T], max_size: int | None = None) -> None:
        if max_size is None:
            max_size = 1
            if cpu_count := os.cpu_count():
                max_size = max(cpu_count, max_size)

        self._item_ctor = item_ctor
        self._pool: list[_T] = []
        self._lock = Lock()
        self.max_size = max_size

    def acquire(self) -> _T:
        with self._lock:
            if self._pool:
                return self._pool.pop(-1)
            # pool is empty, create a new object
            return self._item_ctor()

    def release(self, item: _T) -> None:
        with self._lock:
            if len(self._pool) < self.max_size:
                self._pool.append(item)

    @contextlib.contextmanager
    def item(self) -> Generator[_T, None, None]:
        item = self.acquire()
        try:
            yield item
        finally:
            self.release(item)

    def set_max_size(self, new_size: int) -> None:
        with self._lock:
            if len(self._pool) > new_size:
                self._pool = self._pool[:new_size]
            self.max_size = new_size


class ZStdCompressor(CompressorInterface):
    name = "zstd"
    available = HAS_ZSTD

    COMPRESS_LEVEL = 9

    if HAS_ZSTD:
        _compressor_pool: ClassVar[_Pool[zstandard.ZstdCompressor]] = _Pool(
            lambda: zstandard.ZstdCompressor(level=ZStdCompressor.COMPRESS_LEVEL)
        )
        _decompressor_pool: ClassVar[_Pool[zstandard.ZstdDecompressor]] = _Pool(
            zstandard.ZstdDecompressor
        )

    @classmethod
    def set_pools_max_size(cls, new_size: int) -> None:
        cls._compressor_pool.set_max_size(new_size)
        cls._decompressor_pool.set_max_size(new_size)

    @classmethod
    def dumps(cls, data: bytes) -> bytes:
        with cls._compressor_pool.item() as compressor:
            return compressor.compress(data)

    @classmethod
    def read(cls, data: bytes) -> bytes:
        # data compressed with streaming APIs will not
        # contains the content size in it's frame header
        # which is expected by ZstdDecompressor.decompress
        # use ZstdDecompressionObj instead
        # see: https://github.com/indygreg/python-zstandard/issues/150
        with cls._decompressor_pool.item() as decompressor:
            decompress_obj = decompressor.decompressobj()
            return decompress_obj.decompress(data) + decompress_obj.flush()

    class CompressObj(CompressObjInterface):
        def __init__(self) -> None:
            # zstd compressor is safe to re-use
            # Note that it's not thread safe
            self._compressor: zstandard.ZstdCompressor | None = None
            self._compressobj: zstandard.ZstdCompressionObj | None = None

        def compress(self, data: bytes) -> bytes:
            if self._compressor is None:
                self._compressor = ZStdCompressor._compressor_pool.acquire()
                self._compressobj = self._compressor.compressobj()
            else:
                assert self._compressobj is not None, (
                    "Programming error: _compressobj is None when _compressor is not"
                )

            return self._compressobj.compress(data)

        def flush(self) -> bytes:
            assert self._compressor is not None, (
                "Programming error: Flush called without previous compress"
            )
            assert self._compressobj is not None, (
                "Programming error: _compressobj is None when _compressor is not"
            )

            try:
                return self._compressobj.flush(flush_mode=zstandard.COMPRESSOBJ_FLUSH_FINISH)
            finally:
                # release _compressobj as it's not re-usable
                self._compressobj = None
                # return instance of compressor to pool
                compressor = self._compressor
                self._compressor = None
                ZStdCompressor._compressor_pool.release(compressor)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


from __future__ import annotations

import hmac
import json
import logging
import re
from hashlib import sha1
from typing import TYPE_CHECKING
from typing import Any
from typing import Callable

from dateutil.parser import parse as dateparse
from twisted.internet import defer
from twisted.python import log

from buildbot.process.properties import Properties
from buildbot.util import bytes2unicode
from buildbot.util import httpclientservice
from buildbot.util import unicode2bytes
from buildbot.util.pullrequest import PullRequestMixin
from buildbot.www.hooks.base import BaseHookHandler

if TYPE_CHECKING:
    from twisted.internet.defer import Deferred
    from twisted.web.server import Request

    from buildbot.master import BuildMaster
    from buildbot.util.twisted import InlineCallbacksType

_HEADER_EVENT = b'X-GitHub-Event'
_HEADER_SIGNATURE = b'X-Hub-Signature'

DEFAULT_SKIPS_PATTERN: list[str] = [r'\[ *skip *ci *\]', r'\[ *ci *skip *\]']
DEFAULT_GITHUB_API_URL = 'https://api.github.com'


class GitHubEventHandler(PullRequestMixin):
    property_basename = "github"

    def __init__(
        self,
        secret: str | None,
        strict: bool,
        codebase: str | Callable | None = None,
        github_property_whitelist: list[str] | None = None,
        master: BuildMaster | None = None,
        skips: list[str] | None = None,
        github_api_endpoint: str | None = None,
        pullrequest_ref: str | None = None,
        token: str | None = None,
        debug: bool = False,
        verify: bool = True,
    ):
        if github_property_whitelist is None:
            github_property_whitelist = []
        self._secret = secret
        self._strict = strict
        self._token = token
        self._codebase = codebase
        self.external_property_whitelist = github_property_whitelist
        self.pullrequest_ref = pullrequest_ref
        self.skips = skips
        self.github_api_endpoint = github_api_endpoint
        self.master = master
        if skips is None:
            self.skips = DEFAULT_SKIPS_PATTERN
        if github_api_endpoint is None:
            self.github_api_endpoint = DEFAULT_GITHUB_API_URL

        if self._strict and not self._secret:
            raise ValueError('Strict mode is requested while no secret is provided')
        self.debug = debug
        self.verify = verify

    @defer.inlineCallbacks
    def process(self, request: Request) -> InlineCallbacksType[tuple[list[dict[str, Any]], str]]:
        payload = yield self._get_payload(request)

        event_type = bytes2unicode(request.getHeader(_HEADER_EVENT))
        log.msg(f"X-GitHub-Event: {event_type}", logLevel=logging.DEBUG)

        handler = getattr(self, f'handle_{event_type}', None)

        if handler is None:
            raise ValueError(f'Unknown event: {event_type}')

        result = yield handler(payload, event_type)
        return result

    @defer.inlineCallbacks
    def _get_payload(self, request: Request) -> InlineCallbacksType[dict[str, Any]]:
        assert request.content is not None

        content = bytes2unicode(request.content.read())
        signature = bytes2unicode(request.getHeader(_HEADER_SIGNATURE))

        if not signature and self._strict:
            raise ValueError('Request has no required signature')

        if self._secret and signature:
            try:
                hash_type, hexdigest = signature.split('=')
            except ValueError as e:
                raise ValueError(f'Wrong signature format: {signature}') from e

            if hash_type != 'sha1':
                raise ValueError(f'Unknown hash type: {hash_type}')

            p = Properties()
            p.master = self.master
            rendered_secret = yield p.render(self._secret)

            mac = hmac.new(
                unicode2bytes(rendered_secret), msg=unicode2bytes(content), digestmod=sha1
            )

            def _cmp(a: str, b: str) -> bool:
                return hmac.compare_digest(a, b)

            if not _cmp(bytes2unicode(mac.hexdigest()), hexdigest):
                raise ValueError('Hash mismatch')

        content_type = request.getHeader(b'Content-Type')

        if content_type == b'application/json':
            payload = json.loads(content)
        elif content_type == b'application/x-www-form-urlencoded':
            payload = json.loads(bytes2unicode(request.args[b'payload'][0]))
        else:
            raise ValueError(f'Unknown content type: {content_type}')

        log.msg(f"Payload: {payload}", logLevel=logging.DEBUG)

        return payload

    def handle_ping(self, _: dict[str, Any], __: str) -> tuple[list[dict[str, Any]], str]:
        return [], 'git'

    def handle_workflow_run(self, _: dict[str, Any], __: str) -> tuple[list[dict[str, Any]], str]:
        return [], 'git'

    def handle_push(self, payload: dict[str, Any], event: str) -> tuple[list[dict[str, Any]], str]:
        # This field is unused:
        user: str | None = None
        # user = payload['pusher']['name']
        repo = payload['repository']['name']
        repo_url = payload['repository']['html_url']
        # NOTE: what would be a reasonable value for project?
        # project = request.args.get('project', [''])[0]
        project = payload['repository']['full_name']

        # Inject some additional white-listed event payload properties
        properties = self.extractProperties(payload)
        changes = self._process_change(payload, user, repo, repo_url, project, event, properties)

        log.msg(f"Received {len(changes)} changes from github")

        return changes, 'git'

    @defer.inlineCallbacks
    def handle_pull_request(
        self, payload: dict[str, Any], event: str
    ) -> InlineCallbacksType[tuple[list[dict[str, Any]], str]]:
        changes: list[dict[str, Any]] = []
        number = payload['number']
        refname = f'refs/pull/{number}/{self.pullrequest_ref}'
        basename = payload['pull_request']['base']['ref']
        commits = payload['pull_request']['commits']
        title = payload['pull_request']['title']
        comments = payload['pull_request']['body']
        repo_full_name = payload['repository']['full_name']
        head_sha = payload['pull_request']['head']['sha']
        revlink = payload['pull_request']['_links']['html']['href']

        log.msg(f'Processing GitHub PR #{number}', logLevel=logging.DEBUG)

        head_msg = yield self._get_commit_msg(repo_full_name, head_sha)
        if self._has_skip(head_msg):
            log.msg(f"GitHub PR #{number}, Ignoring: head commit message contains skip pattern")
            return ([], 'git')

        action = payload.get('action')
        if action not in ('opened', 'reopened', 'synchronize'):
            log.msg(f"GitHub PR #{number} {action}, ignoring")
            return (changes, 'git')

        files = yield self._get_pr_files(repo_full_name, number)

        properties = {
            'pullrequesturl': revlink,
            'event': event,
            'basename': basename,
            **self.extractProperties(payload['pull_request']),
        }
        change = {
            'revision': payload['pull_request']['head']['sha'],
            'when_timestamp': dateparse(payload['pull_request']['created_at']),
            'branch': refname,
            'files': files,
            'revlink': payload['pull_request']['_links']['html']['href'],
            'repository': payload['repository']['html_url'],
            'project': payload['pull_request']['base']['repo']['full_name'],
            'category': 'pull',
            # TODO: Get author name based on login id using txgithub module
            'author': payload['sender']['login'],
            'comments': (
                f"GitHub Pull Request #{number} ({commits} "
                f"commit{'s' if commits != 1 else ''})\n{title}\n{comments}"
            ),
            'properties': properties,
        }

        if callable(self._codebase):
            change['codebase'] = self._codebase(payload)
        elif self._codebase is not None:
            change['codebase'] = self._codebase

        changes.append(change)

        log.msg(f"Received {len(changes)} changes from GitHub PR #{number}")
        return (changes, 'git')

    @defer.inlineCallbacks
    def _get_commit_msg(self, repo: str, sha: str) -> InlineCallbacksType[str]:
        """
        :param repo: the repo full name, ``{owner}/{project}``.
            e.g. ``buildbot/buildbot``
        """
        assert self.master is not None

        headers = {
            'User-Agent': 'Buildbot',
        }
        if self._token:
            p = Properties()
            p.master = self.master
            p.setProperty("full_name", repo, "change_hook")
            token = yield p.render(self._token)
            headers['Authorization'] = 'token ' + token

        url = f'/repos/{repo}/commits/{sha}'
        http = yield httpclientservice.HTTPSession(
            self.master.httpservice,
            self.github_api_endpoint,
            headers=headers,
            debug=self.debug,
            verify=self.verify,
        )
        res = yield http.get(url)
        if 200 <= res.code < 300:
            data = yield res.json()
            return data['commit']['message']

        log.msg(f'Failed fetching PR commit message: response code {res.code}')
        return 'No message field'

    @defer.inlineCallbacks
    def _get_pr_files(self, repo: str, number: int) -> InlineCallbacksType[list[str]]:
        """
        Get Files that belong to the Pull Request
        :param repo: the repo full name, ``{owner}/{project}``.
            e.g. ``buildbot/buildbot``
        :param number: the pull request number.
        """
        assert self.master is not None

        headers = {"User-Agent": "Buildbot"}
        if self._token:
            p = Properties()
            p.master = self.master
            p.setProperty("full_name", repo, "change_hook")
            token = yield p.render(self._token)
            headers["Authorization"] = "token " + token

        url = f"/repos/{repo}/pulls/{number}/files"
        http = yield httpclientservice.HTTPSession(
            self.master.httpservice,
            self.github_api_endpoint,
            headers=headers,
            debug=self.debug,
            verify=self.verify,
        )
        res = yield http.get(url)
        if 200 <= res.code < 300:
            data = yield res.json()
            filenames = []
            for f in data:
                filenames.append(f["filename"])
                # If a file was moved this tell us where it was moved from.
                previous_filename = f.get("previous_filename")
                if previous_filename is not None:
                    filenames.append(previous_filename)

            return filenames

        log.msg(f'Failed fetching PR files: response code {res.code}')
        return []

    def _process_change(
        self,
        payload: dict[str, Any],
        user: str | None,
        repo: str,
        repo_url: str,
        project: str,
        event: str,
        properties: dict[str, Any],
    ) -> list[dict[str, Any]]:
        """
        Consumes the JSON as a python object and actually starts the build.

        :arguments:
            payload
                Python Object that represents the JSON sent by GitHub Service
                Hook.
        """
        changes: list[dict[str, Any]] = []
        refname = payload['ref']

        # We only care about regular heads or tags
        match = re.match(r"^refs/(heads|tags)/(.+)$", refname)
        if not match:
            log.msg(f"Ignoring refname `{refname}': Not a branch")
            return changes
        category = None  # None is the legacy category for when hook only supported push
        if match.group(1) == "tags":
            category = "tag"

        branch = match.group(2)
        if payload.get('deleted'):
            log.msg(f"Branch `{branch}' deleted, ignoring")
            return changes

        # check skip pattern in commit message. e.g.: [ci skip] and [skip ci]
        head_msg = payload['head_commit'].get('message', '')
        if self._has_skip(head_msg):
            return changes
        commits = payload['commits']
        if payload.get('created'):
            commits = [payload['head_commit']]
        for commit in commits:
            files = []
            for kind in ('added', 'modified', 'removed'):
                files.extend(commit.get(kind, []))

            when_timestamp = dateparse(commit['timestamp'])

            log.msg(f"New revision: {commit['id'][:8]}")

            change = {
                'author': f"{commit['author']['name']} <{commit['author']['email']}>",
                'committer': f"{commit['committer']['name']} <{commit['committer']['email']}>",
                'files': files,
                'comments': commit['message'],
                'revision': commit['id'],
                'when_timestamp': when_timestamp,
                'branch': branch,
                'revlink': commit['url'],
                'repository': repo_url,
                'project': project,
                'properties': {
                    'github_distinct': commit.get('distinct', True),
                    'event': event,
                },
                'category': category,
            }
            # Update with any white-listed github event properties
            change['properties'].update(properties)

            if callable(self._codebase):
                change['codebase'] = self._codebase(payload)
            elif self._codebase is not None:
                change['codebase'] = self._codebase

            changes.append(change)

        return changes

    def _has_skip(self, msg: str) -> bool:
        """
        The message contains the skipping keyword or not.

        :return type: Bool
        """
        skips = self.skips or []
        for skip in skips:
            if re.search(skip, msg):
                return True
        return False


# for GitHub, we do another level of indirection because
# we already had documented API that encouraged people to subclass GitHubEventHandler
# so we need to be careful not breaking that API.


class GitHubHandler(BaseHookHandler):
    def __init__(self, master: BuildMaster, options: dict[str, Any] | None):
        if options is None:
            options = {}
        super().__init__(master, options)  # type: ignore[arg-type]

        klass = options.get('class', GitHubEventHandler)
        klass_kwargs = {
            'master': master,
            'codebase': options.get('codebase', None),
            'github_property_whitelist': options.get('github_property_whitelist', None),
            'skips': options.get('skips', None),
            'github_api_endpoint': options.get('github_api_endpoint', None)
            or 'https://api.github.com',
            'pullrequest_ref': options.get('pullrequest_ref', None) or 'merge',
            'token': options.get('token', None),
            'debug': options.get('debug', None) or False,
            'verify': options.get('verify', None) or True,
        }
        handler = klass(options.get('secret', None), options.get('strict', False), **klass_kwargs)
        self.handler = handler

    def getChanges(self, request: Request) -> Deferred[tuple[list[dict[str, Any]], str]]:  # type: ignore[override]
        return self.handler.process(request)


github = GitHubHandler



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Portions Copyright Buildbot Team Members
# Portions Copyright Marius Rieder <marius.rieder@durchmesser.ch>


import re

from twisted.internet import defer

from buildbot import config
from buildbot.process import buildstep
from buildbot.process import logobserver


class MockStateObserver(logobserver.LogLineObserver):
    """Supports reading state changes from Mock state.log from mock version
    1.1.23."""

    _line_re = re.compile(r'^.*(Start|Finish): (.*)$')

    def outLineReceived(self, line):
        m = self._line_re.search(line.strip())
        if m:
            if m.group(1) == "Start":
                self.step.descriptionSuffix = [f"[{m.group(2)}]"]
            else:
                self.step.descriptionSuffix = None
            self.step.updateSummary()


class Mock(buildstep.ShellMixin, buildstep.CommandMixin, buildstep.BuildStep):
    """Add the mock logfiles and clean them if they already exist. Add support
    for the root and resultdir parameter of mock."""

    name = "mock"

    renderables = ["root", "resultdir"]

    haltOnFailure = True
    flunkOnFailure = True

    mock_logfiles = ['build.log', 'root.log', 'state.log']

    root = None
    resultdir = None

    def __init__(self, root=None, resultdir=None, **kwargs):
        kwargs = self.setupShellMixin(kwargs, prohibitArgs=['command'])
        super().__init__(**kwargs)
        if root:
            self.root = root
        if resultdir:
            self.resultdir = resultdir

        if not self.root:
            config.error("You must specify a mock root")

        self.command = ['mock', '--root', self.root]
        if self.resultdir:
            self.command += ['--resultdir', self.resultdir]

    @defer.inlineCallbacks
    def run(self):
        # Try to remove the old mock logs first.
        if self.resultdir:
            for lname in self.mock_logfiles:
                self.logfiles[lname] = self.build.path_module.join(self.resultdir, lname)
        else:
            for lname in self.mock_logfiles:
                self.logfiles[lname] = lname
        self.addLogObserver('state.log', MockStateObserver())

        yield self.runRmdir([
            self.build.path_module.join('build', self.logfiles[l]) for l in self.mock_logfiles
        ])

        cmd = yield self.makeRemoteShellCommand()
        yield self.runCommand(cmd)

        return cmd.results()

    def getResultSummary(self):
        self.descriptionSuffix = None
        return super().getResultSummary()


class MockBuildSRPM(Mock):
    """Build a srpm within a mock. Requires a spec file and a sources dir."""

    name = "mockbuildsrpm"

    description = ["mock buildsrpm"]
    descriptionDone = ["mock buildsrpm"]

    spec = None
    sources = '.'

    def __init__(self, spec=None, sources=None, **kwargs):
        """
        Creates the MockBuildSRPM object.

        @type spec: str
        @param spec: the path of the specfiles.
        @type sources: str
        @param sources: the path of the sources dir.
        @type kwargs: dict
        @param kwargs: All further keyword arguments.
        """
        super().__init__(**kwargs)
        if spec:
            self.spec = spec
        if sources:
            self.sources = sources

        if not self.spec:
            config.error("You must specify a spec file")
        if not self.sources:
            config.error("You must specify a sources dir")

        self.command += ['--buildsrpm', '--spec', self.spec, '--sources', self.sources]
        self.addLogObserver('stdio', logobserver.LineConsumerLogObserver(self.logConsumer))

    def logConsumer(self):
        r = re.compile(r"Wrote: .*/([^/]*.src.rpm)")
        while True:
            _, line = yield
            m = r.search(line)
            if m:
                self.setProperty("srpm", m.group(1), 'MockBuildSRPM')


class MockRebuild(Mock):
    """Rebuild a srpm within a mock. Requires a srpm file."""

    name = "mock"

    description = ["mock rebuilding srpm"]
    descriptionDone = ["mock rebuild srpm"]

    srpm = None

    def __init__(self, srpm=None, **kwargs):
        """
        Creates the MockRebuildRPM object.

        @type srpm: str
        @param srpm: the path of the srpm file.
        @type kwargs: dict
        @param kwargs: All further keyword arguments.
        """
        super().__init__(**kwargs)
        if srpm:
            self.srpm = srpm

        if not self.srpm:
            config.error("You must specify a srpm")

        self.command += ['--rebuild', self.srpm]



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


from buildbot.test.util import interfaces


class ConnectionInterfaceTest(interfaces.InterfaceTests):
    def setUp(self):
        # subclasses must set self.conn in this method
        raise NotImplementedError

    def test_sig_notifyOnDisconnect(self):
        @self.assertArgSpecMatches(self.conn.notifyOnDisconnect)
        def notifyOnDisconnect(self, cb):
            pass

    def test_sig_loseConnection(self):
        @self.assertArgSpecMatches(self.conn.loseConnection)
        def loseConnection(self):
            pass

    def test_sig_remotePrint(self):
        @self.assertArgSpecMatches(self.conn.remotePrint)
        def remotePrint(self, message):
            pass

    def test_sig_remoteGetWorkerInfo(self):
        @self.assertArgSpecMatches(self.conn.remoteGetWorkerInfo)
        def remoteGetWorkerInfo(self):
            pass

    def test_sig_remoteSetBuilderList(self):
        @self.assertArgSpecMatches(self.conn.remoteSetBuilderList)
        def remoteSetBuilderList(self, builders):
            pass

    def test_sig_remoteStartCommand(self):
        @self.assertArgSpecMatches(self.conn.remoteStartCommand)
        def remoteStartCommand(self, remoteCommand, builderName, commandId, commandName, args):
            pass

    def test_sig_remoteShutdown(self):
        @self.assertArgSpecMatches(self.conn.remoteShutdown)
        def remoteShutdown(self):
            pass

    def test_sig_remoteStartBuild(self):
        @self.assertArgSpecMatches(self.conn.remoteStartBuild)
        def remoteStartBuild(self, builderName):
            pass

    def test_sig_remoteInterruptCommand(self):
        @self.assertArgSpecMatches(self.conn.remoteInterruptCommand)
        def remoteInterruptCommand(builderName, commandId, why):
            pass



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import os

from parameterized import parameterized
from twisted.internet import defer

from buildbot.process.properties import Interpolate
from buildbot.reporters.http import HttpStatusPush
from buildbot.test.fake.secrets import FakeSecretStorage
from buildbot.test.util.integration import RunMasterBase


class FakeSecretReporter(HttpStatusPush):
    def sendMessage(self, reports):
        assert self.auth == ('user', 'myhttppasswd')
        self.reported = True


class SecretsConfig(RunMasterBase):
    @defer.inlineCallbacks
    def setup_config(self, use_interpolation):
        c = {}
        from buildbot.config import BuilderConfig
        from buildbot.plugins import schedulers
        from buildbot.plugins import steps
        from buildbot.plugins import util
        from buildbot.process.factory import BuildFactory

        fake_reporter = FakeSecretReporter(
            'http://example.com/hook', auth=('user', Interpolate('%(secret:httppasswd)s'))
        )

        c['services'] = [fake_reporter]
        c['schedulers'] = [schedulers.ForceScheduler(name="force", builderNames=["testy"])]

        c['secretsProviders'] = [
            FakeSecretStorage(
                secretdict={"foo": "secretvalue", "something": "more", 'httppasswd': 'myhttppasswd'}
            )
        ]
        f = BuildFactory()

        if use_interpolation:
            if os.name == "posix":
                # on posix we can also check whether the password was passed to the command
                command = Interpolate(
                    'echo %(secret:foo)s | ' + 'sed "s/secretvalue/The password was there/"'
                )
            else:
                command = Interpolate('echo %(secret:foo)s')
        else:
            command = ['echo', util.Secret('foo')]

        f.addStep(steps.ShellCommand(command=command))

        c['builders'] = [BuilderConfig(name="testy", workernames=["local1"], factory=f)]
        yield self.setup_master(c)

        return fake_reporter

    # Note that the secret name must be long enough so that it does not crash with random directory
    # or file names in the build dictionary.
    @parameterized.expand([
        ('with_interpolation', True),
        ('plain_command', False),
    ])
    @defer.inlineCallbacks
    def test_secret(self, name, use_interpolation):
        fake_reporter = yield self.setup_config(use_interpolation)
        build = yield self.doForceBuild(wantSteps=True, wantLogs=True)
        self.assertEqual(build['buildid'], 1)

        # check the command line
        res = yield self.checkBuildStepLogExist(build, "echo <foo>")

        # also check the secrets are replaced in argv
        yield self.checkBuildStepLogExist(build, "argv:.*echo.*<foo>", regex=True)

        # also check that the correct value goes to the command
        if os.name == "posix" and use_interpolation:
            res &= yield self.checkBuildStepLogExist(build, "The password was there")

        self.assertTrue(res)
        # at this point, build contains all the log and steps info that is in the db
        # we check that our secret is not in there!
        self.assertNotIn("secretvalue", repr(build))
        self.assertTrue(fake_reporter.reported)

    @parameterized.expand([
        ('with_interpolation', True),
        ('plain_command', False),
    ])
    @defer.inlineCallbacks
    def test_secretReconfig(self, name, use_interpolation):
        yield self.setup_config(use_interpolation)
        self.master_config_dict['secretsProviders'] = [
            FakeSecretStorage(secretdict={"foo": "different_value", "something": "more"})
        ]

        yield self.master.reconfig()
        build = yield self.doForceBuild(wantSteps=True, wantLogs=True)
        self.assertEqual(build['buildid'], 1)
        res = yield self.checkBuildStepLogExist(build, "echo <foo>")
        self.assertTrue(res)
        # at this point, build contains all the log and steps info that is in the db
        # we check that our secret is not in there!
        self.assertNotIn("different_value", repr(build))


class SecretsConfigPB(SecretsConfig):
    proto = "pb"


class SecretsConfigMsgPack(SecretsConfig):
    proto = "msgpack"



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


import os
import sys
import time
from hashlib import md5

from buildbot.util import unicode2bytes


def tryserver(config):
    jobdir = os.path.expanduser(config["jobdir"])
    job = sys.stdin.read()
    # now do a 'safecat'-style write to jobdir/tmp, then move atomically to
    # jobdir/new . Rather than come up with a unique name randomly, I'm just
    # going to MD5 the contents and prepend a timestamp.
    timestring = f"{time.time()}"
    m = md5()
    job = unicode2bytes(job)
    m.update(job)
    jobhash = m.hexdigest()
    fn = f"{timestring}-{jobhash}"
    tmpfile = os.path.join(jobdir, "tmp", fn)
    newfile = os.path.join(jobdir, "new", fn)
    with open(tmpfile, "wb") as f:
        f.write(job)
    os.rename(tmpfile, newfile)

    return 0



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

import re
from typing import TYPE_CHECKING
from typing import Any

from twisted.web import iweb
from zope.interface import implementer

if TYPE_CHECKING:
    from twisted.web.http import Request

try:
    import brotli
except ImportError:
    brotli = None

try:
    import zstandard
except ImportError:
    zstandard = None  # type: ignore[assignment]


@implementer(iweb._IRequestEncoderFactory)
class _EncoderFactoryBase:
    def __init__(self, encoding_type: bytes, encoder_class: type[_EncoderBase] | None) -> None:
        self.encoding_type = encoding_type
        self.encoder_class = encoder_class
        self.check_regex = re.compile(rb"(:?^|[\s,])" + encoding_type + rb"(:?$|[\s,])")

    def encoderForRequest(self, request: Request) -> _EncoderBase | None:
        """
        Check the headers if the client accepts encoding, and encodes the
        request if so.
        """
        if self.encoder_class is None:
            return None

        acceptHeaders = b",".join(request.requestHeaders.getRawHeaders(b"accept-encoding", []))
        if self.check_regex.search(acceptHeaders):
            src_encodings = request.responseHeaders.getRawHeaders(b"content-encoding")
            if src_encodings:
                encoding = b",".join([*src_encodings, self.encoding_type])
            else:
                encoding = self.encoding_type

            request.responseHeaders.setRawHeaders(b"content-encoding", [encoding])
            return self.encoder_class(request) if self.encoder_class is not None else None
        return None


class BrotliEncoderFactory(_EncoderFactoryBase):
    def __init__(self) -> None:
        super().__init__(b'br', _BrotliEncoder if brotli is not None else None)


class ZstandardEncoderFactory(_EncoderFactoryBase):
    def __init__(self) -> None:
        super().__init__(b'zstd', _ZstdEncoder if zstandard is not None else None)


@implementer(iweb._IRequestEncoder)
class _EncoderBase:
    def __init__(self, request: Request) -> None:
        self._request = request

    def _compress(self, data: bytes) -> bytes:
        return data

    def _flush(self) -> bytes:
        return b''

    def encode(self, data: bytes) -> bytes:
        """
        Write to the request, automatically compressing data on the fly.
        """
        if not self._request.startedWriting:
            # Remove the content-length header, we can't honor it
            # because we compress on the fly.
            self._request.responseHeaders.removeHeader(b"content-length")
        return self._compress(data)

    def finish(self) -> bytes:
        """
        Finish handling the request request, flushing any data from the buffer.
        """
        return self._flush()


class _BrotliEncoder(_EncoderBase):
    def __init__(self, request: Request) -> None:
        super().__init__(request)
        self._compressor = brotli.Compressor() if brotli is not None else None

    def _compress(self, data: bytes) -> bytes:
        if self._compressor is not None:
            return self._compressor.process(data)
        return data

    def _flush(self) -> bytes:
        if self._compressor is not None:
            data = self._compressor.finish()
            self._compressor = None
            return data
        return b''


class _ZstdEncoder(_EncoderBase):
    def __init__(self, request: Request) -> None:
        super().__init__(request)
        self._compressor: Any | None = (
            zstandard.ZstdCompressor(write_content_size=True) if zstandard is not None else None
        )
        self._compressobj: Any | None = (
            self._compressor.compressobj() if self._compressor is not None else None
        )

    def _compress(self, data: bytes) -> bytes:
        if self._compressor is not None:
            return self._compressobj.compress(data) if self._compressobj is not None else data
        return data

    def _flush(self) -> bytes:
        if self._compressor is not None:
            assert self._compressobj is not None
            c_data = self._compressobj.flush()
            self._compressor = None
            self._compressobj = None
            return c_data
        return b''
